# -*- coding: utf-8 -*-
'''
Document Module
'''
from __future__ import division

from collections import OrderedDict
from time import time
from uuid import uuid4

import numpy as np
import pkg_resources
from dipy.utils.optpkg import optional_package

from .fitted_modeling_framework import (
    FittedMultiCompartmentModel,
    FittedMultiCompartmentSphericalMeanModel,
    FittedMultiCompartmentSphericalHarmonicsModel)
from ..optimizers.brute2fine import (
    GlobalBruteOptimizer, Brute2FineOptimizer)
from ..optimizers.mix import MixOptimizer
from ..optimizers.multi_tissue_convex_optimizer import (
    MultiTissueConvexOptimizer)
from ..optimizers_fod.csd_cvxpy import CsdCvxpyOptimizer
from ..optimizers_fod.csd_tournier import CsdTournierOptimizer
from ..utils.spherical_mean import (
    estimate_spherical_mean_multi_shell)
from ..utils.utils import (
    T1_tortuosity,
    parameter_equality,
    fractional_parameter)

pathos, have_pathos, _ = optional_package("pathos")
numba, have_numba, _ = optional_package("numba")
graphviz, have_graphviz, _ = optional_package("graphviz")

if have_graphviz:
    from graphviz import Digraph
if have_pathos:
    import pathos.pools as pp
    from pathos.helpers import cpu_count

GRADIENT_TABLES_PATH = pkg_resources.resource_filename(
    'dmipy', 'data/gradient_tables'
)
SIGNAL_MODELS_PATH = pkg_resources.resource_filename(
    'dmipy', 'signal_models'
)

__all__ = [
    'ModelProperties',
    'MultiCompartmentModelProperties',
    'MultiCompartmentModel',
    'MultiCompartmentSphericalMeanModel',
    'MultiCompartmentSphericalHarmonicsModel',
    'homogenize_x0_to_data',
    'ReturnFixedValue'
]


class ModelProperties:
    "Contains various properties for CompartmentModels."

    S0_response = 1.

    @property
    def parameter_ranges(self):
        """Returns the optimization ranges of the model parameters.
        These ranges are given in O(1) scale so optimization algorithms
        don't suffer from large scale differences in optimization parameters.
        """
        return OrderedDict(self._parameter_ranges.copy())

    @property
    def parameter_scales(self):
        """Returns the optimization scales for the model parameters.
        The scales scale the parameter_ranges to their actual size inside
        optimization algorithms.
        """
        return OrderedDict(self._parameter_scales.copy())

    @property
    def parameter_types(self):
        """Returns the optimization scales for the model parameters.
        The scales scale the parameter_ranges to their actual size inside
        optimization algorithms.
        """
        return OrderedDict(self._parameter_types.copy())

    @property
    def parameter_names(self):
        "Returns the names of model parameters."
        return self._parameter_ranges.keys()

    @property
    def parameter_cardinality(self):
        "Returns the cardinality of model parameters"
        return OrderedDict(
            [(k, len(np.atleast_2d(self.parameter_ranges[k]))) for k in
             self.parameter_ranges])


class MultiCompartmentModelProperties:
    "Class that contains various properties of MultiCompartmentModel instance."

    @property
    def parameter_names(self):
        "Returns the names of model parameters."
        return list(self.parameter_ranges.keys())

    def parameter_vector_to_parameters(self, parameter_vector):
        """Returns the model parameters in dictionary format according to their
        parameter_names. Takes parameter_vector as input, which is the same as
        the output of a FittedMultiCompartmentModel.fitted_parameter_vector.

        Parameters
        ----------
        parameter_vector: array of size (Ndata_x, Ndata_y, ..., Nparameters),
            array that contains the linearized model parameters for an ND-array
            of data voxels.

        Returns
        -------
        parameter: dictionary with parameter_names as parameter keys,
            contains the model parameters in dictionary format.
        """
        parameters = {}
        current_pos = 0
        if parameter_vector.ndim == 1:
            for parameter, card in self.parameter_cardinality.items():
                parameters[parameter] = parameter_vector[
                    current_pos: current_pos + card]
                if card == 1:
                    parameters[parameter] = parameters[parameter][0]
                current_pos += card
        else:
            for parameter, card in self.parameter_cardinality.items():
                parameters[parameter] = parameter_vector[
                    ..., current_pos: current_pos + card]
                if card == 1:
                    parameters[parameter] = parameters[parameter][..., 0]
                current_pos += card
        return parameters

    def parameters_to_parameter_vector(self, **parameters):
        """Returns the model parameters in array format. The input is a
        parameters dictionary that has parameter_names as keys. This is also
        the output of a FittedMultiCompartmentModel.fitted_parameters.

        It's possible to give an array of values for one parameter and only a
        float for others. The function will automatically assume that that the
        float parameters are constant in the data set and broadcast them
        accordingly.

        The output parameter_vector can be used in simulate_data() to generate
        data according to the given input parameters.

        Parameters
        ----------
        parameters: keyword arguments of parameter_names.
            Can be given as **parameter_dictionary that contains the model
            parameter values.

        Returns
        -------
        parameter_vector: array of size (Ndata_x, Ndata_y, ..., Nparameters),
            array that contains the linearized model parameters for an ND-array
            of data voxels.
        """
        parameter_vector = []
        parameter_shapes = []
        for parameter, card in self.parameter_cardinality.items():
            value = np.atleast_1d(parameters[parameter])
            if card == 1 and not np.all(value.shape == np.r_[1]):
                parameter_shapes.append(value.shape)
            elif card > 1 and not np.all(value.shape == np.r_[card]):
                parameter_shapes.append(value.shape[:-1])

        if len(set(parameter_shapes)) > 1:
            msg = "parameter shapes are inconsistent."
            raise ValueError(msg)
        elif len(set(parameter_shapes)) == 0:
            for parameter, card in self.parameter_cardinality.items():
                parameter_vector.append(parameters[parameter])
            parameter_vector = np.hstack(parameter_vector)
        elif len(set(parameter_shapes)) == 1:
            for parameter, card in self.parameter_cardinality.items():
                value = np.atleast_1d(parameters[parameter])
                if card == 1 and np.all(value.shape == np.r_[1]):
                    parameter_vector.append(
                        np.tile(value[0], np.r_[parameter_shapes[0], 1]))
                elif card == 1 and not np.all(value.shape == np.r_[1]):
                    parameter_vector.append(value[..., None])
                elif card > 1 and np.all(value.shape == np.r_[card]):
                    parameter_vector.append(
                        np.tile(value, np.r_[parameter_shapes[0], 1])
                    )
                else:
                    parameter_vector.append(parameters[parameter])
            parameter_vector = np.concatenate(parameter_vector, axis=-1)
        return parameter_vector

    def parameter_initial_guess_to_parameter_vector(self, **parameters):
        """Function that returns a parameter_vector while allowing for partial
        input of model parameters, setting the ones that were not given to
        'None'. Such an array can be given to the fit() function to provide an
        initial parameter guess when fitting the data to the model.

        Parameters
        ----------
        parameters: keyword arguments of parameter names,
            parameter values of only the parameters you want to give as an
            initial condition for the optimizer.

        Returns
        -------
        parameter_vector: array of size (Ndata_x, Ndata_y, ..., Nparameters),
            array that contains the linearized model parameters for an ND-array
            of data voxels, with None's for non-given parameters.
        """
        set_parameters = {}
        parameter_cardinality = self.parameter_cardinality.copy()
        for parameter, value in parameters.items():
            if parameter in self.parameter_cardinality.keys():
                set_parameters[parameter] = value
                del parameter_cardinality[parameter]
            else:
                msg = '"{}" is not a valid model parameter.'.format(parameter)
                raise ValueError(msg)
        if len(parameter_cardinality) == 0:
            print("All model parameters set or have initial guess.")
        else:
            for parameter, card in parameter_cardinality.items():
                set_parameters[parameter] = np.tile(np.nan, card)
        return self.parameters_to_parameter_vector(**set_parameters)

    def _prepare_parameters(self):
        """Prepares the parameter ranges, scales, cadinality and parameter
        upon instantiating the MultiCompartmentModel"""
        self.model_names = []
        model_counts = {}

        for model in self.models:
            if model.__class__ not in model_counts:
                model_counts[model.__class__] = 1
            else:
                model_counts[model.__class__] += 1

            self.model_names.append(
                '{}_{:d}_'.format(
                    model.__class__.__name__,
                    model_counts[model.__class__]
                )
            )

        self.parameter_ranges = OrderedDict({
            model_name + k: v
            for model, model_name in zip(self.models, self.model_names)
            for k, v in model.parameter_ranges.items()
        })

        self.parameter_scales = OrderedDict({
            model_name + k: v
            for model, model_name in zip(self.models, self.model_names)
            for k, v in model.parameter_scales.items()
        })

        self.parameter_types = OrderedDict({
            model_name + k: v
            for model, model_name in zip(self.models, self.model_names)
            for k, v in model.parameter_types.items()
        })

        self._parameter_map = {
            model_name + k: (model, k)
            for model, model_name in zip(self.models, self.model_names)
            for k in model.parameter_ranges
        }

        self._inverted_parameter_map = {
            v: k for k, v in self._parameter_map.items()
        }

        self.parameter_cardinality = OrderedDict([
            (k, len(np.atleast_2d(self.parameter_ranges[k])))
            for k in self.parameter_ranges
        ])

    def _prepare_partial_volumes(self):
        "Prepares partial volumes upon instantiating the MultiCompartmentModel"
        if len(self.models) > 1:
            self.partial_volume_names = [
                'partial_volume_{:d}'.format(i)
                for i in range(len(self.models))
            ]

            for i, partial_volume_name in enumerate(
                    self.partial_volume_names):
                self.parameter_ranges[partial_volume_name] = (0.01, .99)
                self.parameter_scales[partial_volume_name] = 1.
                self._parameter_map[partial_volume_name] = (
                    None, partial_volume_name
                )
                self.parameter_types[partial_volume_name] = 'normal'
                self._inverted_parameter_map[(None, partial_volume_name)] = \
                    partial_volume_name
                self.parameter_cardinality[partial_volume_name] = 1
        else:
            self.partial_volume_names = []

    def _prepare_parameter_links(self):
        """Prepares parameter links if given as input to MultiCompartmentModel.
        It first checks if the parameter that will be linked exists. If so,
        then it removes it from the parameter ranges, scales and cardinality,
        so it will not be optimized (as it will be a function of other
        parameters)."""
        for i, parameter_function in enumerate(self.parameter_links):
            parameter_model, parameter_name, parameter_function, arguments = \
                parameter_function

            if (
                    (parameter_model, parameter_name)
                    not in self._inverted_parameter_map
            ):
                raise ValueError(
                    "Parameter function {} doesn't exist".format(i)
                )

            parameter_name = self._inverted_parameter_map[
                (parameter_model, parameter_name)
            ]

            del self.parameter_ranges[parameter_name]
            del self.parameter_cardinality[parameter_name]
            del self.parameter_scales[parameter_name]
            del self.parameter_types[parameter_name]
            del self.parameter_optimization_flags[parameter_name]

    def _prepare_model_properties(self):
        """Checks that spherical mean and regular models cannot be optimized
        together, and whether the model can estimate a Fiber Orientation
        Distribution (FOD)."""
        self.fod_available = False
        for model in self.models:
            try:
                model.fod
                self.fod_available = True
            except AttributeError:
                pass

    def _check_for_double_model_class_instances(self):
        "Checks all models have unique class instances."
        if len(self.models) != len(set(self.models)):
            msg = "Each model in the multi-compartment model must be "
            msg += "instantiated separately. For example, to make a model "
            msg += "with two sticks, the models must be given as "
            msg += "models = [stick1, stick2], not as "
            msg += "models = [stick1, stick1]."
            raise ValueError(msg)

    def add_linked_parameters_to_parameters(self, parameters):
        """When making the MultiCompartmentModel function call, adds the linked
        parameter to the optimized parameters by evaluating the parameter link
        function."""
        if len(self.parameter_links) == 0:
            return parameters
        parameters = parameters.copy()
        for parameter in self.parameter_links[::-1]:
            parameter_model, parameter_name, parameter_function, arguments = \
                parameter
            parameter_name = self._inverted_parameter_map[
                (parameter_model, parameter_name)
            ]

            if len(arguments) > 0:
                argument_values = []
                for argument in arguments:
                    argument_name = self._inverted_parameter_map[argument]
                    argument_values.append(parameters.get(
                        argument_name
                    ))

                parameters[parameter_name] = parameter_function(
                    *argument_values
                )
            else:
                parameters[parameter_name] = parameter_function()
        return parameters

    def _check_if_volume_fractions_are_fixed(self):
        "checks if volume fractions have been fixed."
        self.volume_fractions_fixed = True
        for name, flag in self.parameter_optimization_flags.items():
            if flag and name in self.partial_volume_names:
                self.volume_fractions_fixed = False

    def _prepare_parameters_to_optimize(self):
        "Sets up which parmameters to optimize."
        self.parameter_optimization_flags = OrderedDict({
            k: True
            for k, v in self.parameter_cardinality.items()
        })

    @property
    def bounds_for_optimization(self):
        "Returns the linear parameter bounds for the model optimization."
        bounds = []
        for parameter, card in self.parameter_cardinality.items():
            range_ = self.parameter_ranges[parameter]
            if card == 1:
                bounds.append(range_)
            else:
                for i in range(card):
                    bounds.append((range_[i][0], range_[i][1]))
        return bounds

    @property
    def opt_params_for_optimization(self):
        "Returns the linear bools whether to optimize a model parameter."
        params = []
        for parameter, card in self.parameter_cardinality.items():
            optimize_param = self.parameter_optimization_flags[parameter]
            if card == 1:
                params.append(optimize_param)
            else:
                for i in range(card):
                    params.append(optimize_param)
        return params

    @property
    def scales_for_optimization(self):
        "Returns the linear parameter scales for model optimization."
        return np.hstack([scale for parameter, scale in
                          self.parameter_scales.items()])

    def _check_for_tortuosity_constraint(self):
        for link in self.parameter_links:
            if isinstance(link[2], T1_tortuosity):
                msg = "Cannot use MIX optimization when the Tortuosity "
                msg += "constraint is set in the MultiCompartmentModel. To "
                msg += "use MIX while imposing Tortuosity, set the constraint "
                msg += "in the DistributedModel step."
                raise ValueError(msg)

    def set_initial_guess_parameter(self, parameter_name, value):
        """
        Allows the user to fix an optimization parameter to a static value.
        The fixed parameter will be removed from the optimized parameters and
        added as a linked parameter.

        Parameters
        ----------
        parameter_name: string
            name of the to-be-fixed parameters, see self.parameter_names.
        value: float or list of corresponding parameter_cardinality.
            the value to fix the parameter at in SI units.
        """
        if parameter_name in self.parameter_ranges.keys():
            card = self.parameter_cardinality[parameter_name]
            if card == 1:
                if isinstance(value, int) or isinstance(value, float):
                    self.x0_parameters[parameter_name] = value
                elif isinstance(value, np.ndarray):
                    self._add_initial_guess_parameter_array(
                        parameter_name, value)
            elif card >= 2:
                value = np.array(value, dtype=float)
                if value.shape[-1] != card:
                    msg = '{} can only be fixed to an array or list with ' \
                          'last dimension {}.'
                    raise ValueError(msg.format(parameter_name, type(value)))
                if value.ndim == 1:
                    self.x0_parameters[parameter_name] = value
                if value.ndim > 1:
                    self._add_initial_guess_parameter_array(
                        parameter_name, value)
        else:
            msg = '{} does not exist or has already been fixed.'.format(
                parameter_name)
            raise ValueError(msg)

    def _add_initial_guess_parameter_array(
            self, parameter_name, parameter_array):
        temp_dict = self.x0_parameters.copy()
        temp_dict[parameter_name] = parameter_array
        try:
            self.parameter_initial_guess_to_parameter_vector(
                **temp_dict)
            self.x0_parameters = temp_dict
        except ValueError:
            msg = '{} does not have the same shape'.format(parameter_name)
            msg += 'as the previously fixed parameters.'
            raise ValueError(msg)

    def set_fixed_parameter(self, parameter_name, value):
        """
        Allows the user to fix an optimization parameter to a static value.
        The fixed parameter will be removed from the optimized parameters and
        added as a linked parameter.

        Parameters
        ----------
        parameter_name: string
            name of the to-be-fixed parameters, see self.parameter_names.
        value: float or list of corresponding parameter_cardinality.
            the value to fix the parameter at in SI units.
        """
        if parameter_name in self.parameter_ranges.keys():
            card = self.parameter_cardinality[parameter_name]
            if card == 1:
                if isinstance(value, int) or isinstance(value, float):
                    self._add_fixed_parameter_value(parameter_name,
                                                    float(value))
                elif isinstance(value, np.ndarray):
                    self._add_fixed_parameter_array(parameter_name, value)
                else:
                    msg = 'fixed value for {} must be number or np.array, ' \
                          'currently {}'
                    raise ValueError(msg.format(parameter_name, type(value)))
            elif card >= 2:
                value = np.array(value, dtype=float)
                if value.shape[-1] != card:
                    msg = '{} can only be fixed to an array or list with ' \
                          'last dimension {}.'
                    raise ValueError(msg.format(parameter_name, card))
                if value.ndim == 1:
                    self._add_fixed_parameter_value(parameter_name, value)
                if value.ndim > 1:
                    self._add_fixed_parameter_array(parameter_name, value)
        else:
            msg = '{} does not exist or has already been fixed.'.format(
                parameter_name)
            raise ValueError(msg)

    def _add_fixed_parameter_value(self, parameter_name, value):
        model, name = self._parameter_map[parameter_name]
        parameter_link = (model, name, ReturnFixedValue(value), [])
        self.parameter_links.append(parameter_link)
        del self.parameter_ranges[parameter_name]
        del self.parameter_cardinality[parameter_name]
        del self.parameter_scales[parameter_name]
        del self.parameter_types[parameter_name]
        del self.parameter_optimization_flags[parameter_name]

    def _add_fixed_parameter_array(self, parameter_name, parameter_array):
        temp_dict = self.x0_parameters.copy()
        temp_dict[parameter_name] = parameter_array
        try:
            self.parameter_initial_guess_to_parameter_vector(
                **temp_dict)
            self.x0_parameters = temp_dict
            self.parameter_optimization_flags[parameter_name] = False
        except ValueError:
            msg = '{} does not have the same shape'.format(parameter_name)
            msg += 'as the previously fixed parameters.'
            raise ValueError(msg)

    def set_tortuous_parameter(self, lambda_perp_parameter_name,
                               lambda_par_parameter_name,
                               volume_fraction_intra_parameter_name,
                               volume_fraction_extra_parameter_name,
                               S0_correction=False):
        """
        Allows the user to set a tortuosity constraint on the perpendicular
        diffusivity of the extra-axonal compartment, which depends on the
        intra-axonal volume fraction and parallel diffusivity.

        The perpendicular diffusivity parameter will be removed from the
        optimized parameters and added as a linked parameter.

        To employ the multi-tissue correction of tortuosity it is sufficient to
        pass the S0_intra and S0_extra parameters.

        Parameters
        ----------
        lambda_perp_parameter_name: string
            name of the perpendicular diffusivity parameter, see
            self.parameter_names.
        lambda_par_parameter_name: string
            name of the parallel diffusivity parameter, see
            self.parameter_names.
        volume_fraction_intra_parameter_name: string
            name of the intra-axonal volume fraction parameter, see
            self.parameter_names.
        volume_fraction_extra_parameter_name: string
            name of the extra-axonal volume fraction parameter, see
            self.parameter_names.
        S0_correction: bool
            If True, it uses the S0 of the intra-axonal and extra-axonal
            compartments to define the tortuosity constraint. Default: False.
        """
        params = [lambda_perp_parameter_name, lambda_par_parameter_name,
                  volume_fraction_intra_parameter_name,
                  volume_fraction_extra_parameter_name]
        for param in params:
            try:
                self.parameter_cardinality[param]
            except KeyError:
                msg = ("{} does not exist or has already been fixed.").format(
                    param)
                raise ValueError(msg)

        model, name = self._parameter_map[lambda_perp_parameter_name]
        if S0_correction and self.S0_tissue_responses is not None:
            s0intra_tag = volume_fraction_intra_parameter_name.split('_')[-1]
            s0extra_tag = volume_fraction_extra_parameter_name.split('_')[-1]
            S0_intra = self.S0_tissue_responses[int(s0intra_tag)]
            S0_extra = self.S0_tissue_responses[int(s0extra_tag)]
            print('Employing S0 correction of tortuosity constraint with:')
            print('S0_intra: {}'.format(S0_intra))
            print('S0_extra: {}'.format(S0_extra))
        else:
            S0_intra = 1.
            S0_extra = 1.
        tortuosity = T1_tortuosity(S0_intra, S0_extra)

        self.parameter_links.append([model, name, tortuosity, [
            self._parameter_map[lambda_par_parameter_name],
            self._parameter_map[volume_fraction_intra_parameter_name],
            self._parameter_map[volume_fraction_extra_parameter_name]]])
        del self.parameter_ranges[lambda_perp_parameter_name]
        del self.parameter_cardinality[lambda_perp_parameter_name]
        del self.parameter_scales[lambda_perp_parameter_name]
        del self.parameter_types[lambda_perp_parameter_name]
        del self.parameter_optimization_flags[lambda_perp_parameter_name]

    def set_equal_parameter(self, parameter_name_in, parameter_name_out):
        """
        Allows the user to set two parameters equal to each other. This is used
        for example in the NODDI model to set the parallel diffusivities of the
        Stick and Zeppelin compartment to the same value.

        The second input parameter will be removed from the optimized
        parameters and added as a linked parameter.

        Parameters
        ----------
        parameter_name_in: string
            the first parameter name, see self.parameter_names.
        parameter_name_out: string,
            the second parameter name, see self.parameter_names. This is the
            parameter that will be removed form the optimzed parameters.
        """
        params = [parameter_name_in, parameter_name_out]
        for param in params:
            try:
                self.parameter_cardinality[param]
            except KeyError:
                msg = ("{} does not exist or has already been fixed.").format(
                    param)
                raise ValueError(msg)
        model, name = self._parameter_map[parameter_name_out]
        self.parameter_links.append([model, name, parameter_equality, [
            self._parameter_map[parameter_name_in]]])
        del self.parameter_ranges[parameter_name_out]
        del self.parameter_cardinality[parameter_name_out]
        del self.parameter_scales[parameter_name_out]
        del self.parameter_types[parameter_name_out]
        del self.parameter_optimization_flags[parameter_name_out]

    def set_fractional_parameter(self,
                                 parameter1_smaller_equal_than, parameter2):
        r"""
        Allows to impose a constraint to make one parameter smaller or equal to
        another parameter. This is done by replacing parameter1 with a
        new parameter that is defined as a fraction between 0 and 1 of
        parameter2. The new parameter will be the same as the old parameter
        name with "_fraction" appended to it.

        Parameters
        ----------
        parameter1_smaller_equal_than: string
            parameter name to be made a fraction of parameter2
        parameter2: string
            the parameter that is larger or equal than parameter1
        """
        params = [parameter1_smaller_equal_than, parameter2]
        for param in params:
            try:
                self.parameter_cardinality[param]
            except KeyError:
                msg = ("{} does not exist or has already been fixed.").format(
                    param)
                raise ValueError(msg)
        # append new parameter to parameters
        new_parameter_name = parameter1_smaller_equal_than + '_fraction'

        self._add_optimization_parameter(
            new_parameter_name, [0., 1.], 1., 1, 'normal', True)
        model, name = self._parameter_map[parameter1_smaller_equal_than]
        self.parameter_links.append([model, name, fractional_parameter, [
            self._parameter_map[new_parameter_name],
            self._parameter_map[parameter2]]])

        # remove old parameter1
        del self.parameter_ranges[parameter1_smaller_equal_than]
        del self.parameter_cardinality[parameter1_smaller_equal_than]
        del self.parameter_scales[parameter1_smaller_equal_than]
        del self.parameter_types[parameter1_smaller_equal_than]
        del self.parameter_optimization_flags[parameter1_smaller_equal_than]

    def _add_optimization_parameter(
            self,
            parameter_name,
            parameter_range,
            parameter_scale,
            parameter_card,
            parameter_type,
            parameter_flag):
        """
        Creates new ordered dictionaries for model properties with the
        optimization parameter on top.
        """
        old_parameter_ranges = self.parameter_ranges
        old_parameter_scales = self.parameter_scales
        old_parameter_cardinality = self.parameter_cardinality
        old_parameter_types = self.parameter_types
        old_optimization_flags = self.parameter_optimization_flags

        self.parameter_ranges = OrderedDict({parameter_name: parameter_range})
        self.parameter_scales = OrderedDict({parameter_name: parameter_scale})
        self.parameter_cardinality = OrderedDict(
            {parameter_name: parameter_card})
        self.parameter_types = OrderedDict({parameter_name: parameter_type})
        self.parameter_optimization_flags = OrderedDict(
            {parameter_name: parameter_flag})

        for name, _ in old_parameter_ranges.items():
            self.parameter_ranges.update({name: old_parameter_ranges[name]})
            self.parameter_scales.update({name: old_parameter_scales[name]})
            self.parameter_cardinality.update(
                {name: old_parameter_cardinality[name]})
            self.parameter_types.update({name: old_parameter_types[name]})
            self.parameter_optimization_flags.update(
                {name: old_optimization_flags[name]})

        self._parameter_map.update({parameter_name: (None, 'fraction')})
        self._inverted_parameter_map.update(
            {(None, 'fraction'): parameter_name})

    def _check_model_params_with_acquisition_params(self, acquisition_scheme):
        for model in self.models:
            for parameter in model._required_acquisition_parameters:
                if getattr(acquisition_scheme, parameter) is None:
                    msg = "{} is not compatible with ".format(
                        model.__class__.__name__)
                    msg += "given acquisition scheme because it needs "
                    msg += "{} as an acquisition parameter.".format(parameter)
                    raise ValueError(msg)

    def visualize_model_setup(
            self, view=True, cleanup=True, with_parameters=False,
            im_format='png'):
        """
        Visualizes MultiCompartmentModel setup using graphviz module. It uses
        the uuid module to create a unique identifier for each model in the
        MultiCompartmentModel to make sure each node is referenced in a unique
        way.

        If cleanup is set to False it will save the PDF of the graph in the
        current working directory.

        If with_parameters is set to true, it will include all the parameters
        of each model in the graph. Note the graph will ignore any parameter
        links that may have already been imposed (e.g. parameter equality or
        fixed parameters).

        Parameters
        ----------
        view: boolean,
            Whether or not to visualize the graph in a popup screen.
        cleanup: boolean,
            Whether or not to delete the PDF file of the model setup.
        with_parameters: boolean,
            Whether or not to also visualize the parameters of each model.
        """
        if not have_graphviz:
            raise ImportError('graphviz package not installed.')
        dot = Digraph('Model Setup', format=im_format)
        base_model = self.__class__.__name__
        base_uuid = str(uuid4())
        dot.node(base_uuid, base_model)
        self._add_recursive_graph_node(dot, base_uuid, self, with_parameters)
        dot.render('Model Setup', view=view, cleanup=cleanup)

    def _add_recursive_graph_node(
            self, graph_model, entry_uuid, entry_model, with_parameters):
        """
        Recursive function to visualize model setup. For every model in a
        MultiCompartmentModel or a distributed model it will check if it is
        a distribution, in which case the function will call itself with the
        sub-model as input and continue until it has found the bottom of the
        model setup.

        Parameters
        ----------
        graph_model: graphviz model instance,
            Instantiated model instance to keep growing with nodes.
        entry_uuid: string,
            Entry model unique identifier from which to keep growing the graph.
        entry_model: dmipy model instance,
            Entry dmipy model from which to keep growing the graph.
        """
        for sub_model in entry_model.models:
            model_name = sub_model.__class__.__name__
            model_uuid = str(uuid4())
            graph_model.node(model_uuid, model_name)
            graph_model.edge(model_uuid, entry_uuid)
            if (sub_model._model_type == 'SphericalDistributedModel' or
                    sub_model._model_type == 'SpatialDistributedModel' or
                    sub_model._model_type == 'BundleModel'):
                self._add_recursive_graph_node(
                    graph_model, model_uuid, sub_model, with_parameters)
            elif with_parameters:
                self._add_parameter_nodes(graph_model, model_uuid, sub_model)
        if hasattr(entry_model, 'distribution'):
            dist_name = entry_model.distribution.__class__.__name__
            dist_uuid = str(uuid4())
            graph_model.node(dist_uuid, dist_name)
            graph_model.edge(dist_uuid, entry_uuid)
            if with_parameters:
                self._add_parameter_nodes(
                    graph_model, dist_uuid, entry_model.distribution)

    def _add_parameter_nodes(self, graph_model, entry_uuid, entry_model):
        """
        Adds the parameters to the graph truee if with_parameters=True in the
        visualize_model_setup function.

        Parameters
        ----------
        graph_model: graphviz model instance,
            Instantiated model instance to keep growing with nodes.
        entry_uuid: string,
            Entry model unique identifier from which to keep growing the graph.
        entry_model: dmipy model instance,
            Entry dmipy model from which to keep growing the graph.
        """
        for parameter_name in entry_model.parameter_names:
            parameter_uuid = str(uuid4())
            graph_model.node(parameter_uuid, parameter_name)
            graph_model.edge(parameter_uuid, entry_uuid)

    def _check_tissue_model_acquisition_scheme(self, acquisition_scheme):
        """Tests if acquisition scheme between MC-model and tissue response
        model are the same.

        Parameters
        ----------
        acquisition_scheme : DmipyAcquisitionScheme instance,
            An acquisition scheme that has been instantiated using Dmipy.
        """
        for model in self.models:
            if model._model_type == 'TissueResponseModel':
                mc_scheme_params = [
                    acquisition_scheme.shell_bvalues,
                    acquisition_scheme.shell_delta,
                    acquisition_scheme.shell_Delta,
                    acquisition_scheme.shell_gradient_strengths]
                tr_scheme_params = [
                    model.acquisition_scheme.shell_bvalues,
                    model.acquisition_scheme.shell_delta,
                    model.acquisition_scheme.shell_Delta,
                    model.acquisition_scheme.shell_gradient_strengths]
                try:
                    np.testing.assert_array_almost_equal(
                        mc_scheme_params, tr_scheme_params)
                except AssertionError:
                    msg = "Acquisition scheme of MC-model and tissue response "
                    msg += "model are not the same."
                    raise ValueError(msg)

    def _check_acquisition_scheme_has_b0s(self, acquisition_scheme):
        """
        Checks if acquisition scheme has any b0-measurements. This for the
        moment is a prerequisite for signal-attenuation-based model fitting.

        Parameters
        ----------
        acquisition_scheme : DmipyAcquisitionScheme instance,
            An acquisition scheme that has been instantiated using Dmipy.
        """
        if not np.any(acquisition_scheme.b0_mask):
            raise ValueError('acquisition scheme must have b0-measurements '
                             'for signal-attenuation-based model fitting.')

    def _construct_convolution_kernel(self, acquisition_scheme=None, **kwargs):
        """
        Helper function that constructs the convolution kernel for the given
        multi-compartment model and the initial condition x0_vector.

        First the parameter vector is converted to a dictionary with the
        corresponding parameter names. Then, the linked parameters are added to
        the given ones. Finally, the rotational harmonics of the model is
        passed to the construct_model_based_A_matrix, which constructs the
        kernel for an arbitrary PGSE-acquisition scheme.

        For multiple models with fixed volume fractions, the A-matrices
        are combined to have a combined convolution kernel.

        For multiple models without fixed volume fractions, the convolution
        kernels for anisotropic and isotropic models are concatenated, with
        the isotropic kernels always having a spherical harmonics order of 0.

        Parameters
        ----------
        acquisition_scheme: DmipyAcquisitionScheme instance,
            An acquisition scheme that has been instantiated using Dmipy.

        Returns
        -------
        kernel: array of size (N_coef, N_data),
            Observation matrix that maps the FOD spherical harmonics
            coefficients to the DWI signal values.
        """
        parameters_dict = self.add_linked_parameters_to_parameters(kwargs)

        if acquisition_scheme is None:
            acquisition_scheme = self.scheme

        if self.volume_fractions_fixed:
            if self.N_models > 1:
                partial_volumes = [
                    parameters_dict[p] for p in self.partial_volume_names
                ]
            else:
                partial_volumes = [1.]
            kernel = 0.
            for model, partial_volume, S0 in zip(self.models,
                                                 partial_volumes,
                                                 self.S0_responses):
                parameters = {}
                for parameter in model.parameter_ranges:
                    parameter_name = self._inverted_parameter_map[
                        (model, parameter)
                    ]
                    parameters[parameter] = parameters_dict.get(
                        parameter_name
                    )
                kernel += S0 * partial_volume * (
                    model.convolution_kernel_matrix(
                        acquisition_scheme, self.sh_order, **parameters))
        else:
            kernel = []
            for model, S0 in zip(self.models, self.S0_responses):
                parameters = {}
                for parameter in model.parameter_ranges:
                    parameter_name = self._inverted_parameter_map[
                        (model, parameter)
                    ]
                    parameters[parameter] = parameters_dict.get(
                        parameter_name
                    )
                if 'orientation' in model.parameter_types.values():
                    kernel.append(S0 * model.convolution_kernel_matrix(
                        acquisition_scheme, self.sh_order, **parameters))
                else:
                    kernel.append(S0 * model.convolution_kernel_matrix(
                        acquisition_scheme, 0, **parameters))

            kernel = np.hstack(kernel)
        return kernel

    def set_parameter_optimization_bounds(self, parameter_name, bounds):
        """
        Sets the parameter optimization bounds for a given parameter.

        Parameters
        ----------
        parameter_name: string,
            name of the parameter whose bounds should be changed.
        bounds: array or size(card, 2),
            upper and lower bound for each optimized value for the given
            parameter, where card is
            self.parameter_cardinality[parameter_name]).

        Raises
        ------
        ValueError: parameter name not in model parameters
        ValueError: input bounds are not of correct shape [card, 2]
        ValueError: input higher bound is lower than lower bound
        """
        if parameter_name not in self.parameter_names:
            raise ValueError(
                '{} not in model parameters'.format(parameter_name))
        card = self.parameter_cardinality[parameter_name]
        bounds_array = np.atleast_2d(bounds)
        input_card, N_bounds = bounds_array.shape[:2]
        if bounds_array.ndim > 2 or input_card != card or N_bounds != 2:
            msg = '{} bounds must be of shape ({}, 2), currently {}.'
            raise ValueError(
                msg.format(parameter_name, card, bounds_array.shape))
        for lower, higher in bounds_array:
            if higher < lower:
                msg = 'given optimization bounds for {} are invalid: lower ' \
                      'bound {} is higher than upper bound {}.'
                raise ValueError(msg.format(parameter_name, lower, higher))
        parameter_scale = np.max(bounds)
        ranges = np.array(bounds) / parameter_scale
        self.parameter_ranges[parameter_name] = ranges
        self.parameter_scales[parameter_name] = parameter_scale


class MultiCompartmentModel(MultiCompartmentModelProperties):
    r'''
    The MultiCompartmentModel class allows to combine any number of
    CompartmentModels and DistributedModels into one combined model that can
    be used to fit and simulate dMRI data.

    Parameters
    ----------
    models : list of N CompartmentModel instances,
        the models to combine into the MultiCompartmentModel.
    parameter_links : list of iterables (model, parameter name, link function,
        argument list),
        deprecated, for testing only.
    '''

    def __init__(self, models, S0_tissue_responses=None, parameter_links=None):
        self.models = models
        self.N_models = len(models)
        if S0_tissue_responses is not None:
            if len(S0_tissue_responses) != self.N_models:
                msg = 'Number of S0_tissue responses {} must be same as ' \
                      'number of input models {}.'
                raise ValueError(
                    msg.format(len(S0_tissue_responses), self.N_models))
        self.S0_tissue_responses = S0_tissue_responses
        self.parameter_links = parameter_links
        if parameter_links is None:
            self.parameter_links = []

        self._prepare_parameters()
        self._prepare_partial_volumes()
        self._prepare_parameter_links()
        self._prepare_model_properties()
        self._check_for_double_model_class_instances()
        self._prepare_parameters_to_optimize()
        self._check_for_NMR_and_other_models()
        self.x0_parameters = {}

        if not have_numba:
            msg = "We highly recommend installing numba for faster function "
            msg += "execution and model fitting."
            print(msg)
        if not have_pathos:
            msg = "We highly recommend installing pathos to take advantage of "
            msg += "multicore processing."
            print(msg)

    def _check_for_NMR_and_other_models(self):
        model_types = [model._model_type for model in self.models]
        if "NMRModel" in model_types:
            if len(np.unique(model_types)) > 1:
                msg = "Cannot combine 1D-NMR and other 3D model types together"
                msg += " into a MultiCompartmentModel."
                raise ValueError(msg)

    def _check_if_sh_coeff_fixed_if_present(self):
        msg = 'sh_coeff parameter {} must be fixed in standard MC models ' \
              'to estimate the kernel parameters.'
        for name, par_type in self.parameter_types.items():
            if par_type == 'sh_coefficients':
                if self.parameter_optimization_flags[name]:
                    raise ValueError(msg.format(name))

    def fit(self, acquisition_scheme, data,
            mask=None, solver='brute2fine', Ns=5, maxiter=300,
            N_sphere_samples=30, use_parallel_processing=have_pathos,
            number_of_processors=None):
        """ The main data fitting function of a MultiCompartmentModel.

        This function can fit it to an N-dimensional dMRI data set, and returns
        a FittedMultiCompartmentModel instance that contains the fitted
        parameters and other useful functions to study the results.

        No initial guess needs to be given to fit a model, but a partial or
        complete initial guess can be given if the user wants to have a
        solution that is a local minimum close to that guess. The
        parameter_initial_guess input can be created using
        parameter_initial_guess_to_parameter_vector().

        A mask can also be given to exclude voxels from fitting (e.g. voxels
        that are outside the brain). If no mask is given then all voxels are
        included.

        An optimization approach can be chosen as either 'brute2fine' or 'mix'.
        - Choosing brute2fine will first use a brute-force optimization to find
          an initial guess for parameters without one, and will then refine the
          result using gradient-descent-based optimization.

          Note that given no initial guess will make brute2fine precompute an
          global parameter grid that will be re-used for all voxels, which in
          many cases is much faster than giving voxel-varying initial condition
          that requires a grid to be estimated per voxel.

        - Choosing mix will use the recent MIX algorithm based on separation of
          linear and non-linear parameters. MIX first uses a stochastic
          algorithm to find the non-linear parameters (non-volume fractions),
          then estimates the volume fractions while fixing the estimates of the
          non-linear parameters, and then finally refines the solution using
          a gradient-descent-based algorithm.

        The fitting process can be readily parallelized using the optional
        "pathos" package. If it is installed then it will automatically use it,
        but it can be turned off by setting use_parallel_processing=False. The
        algorithm will automatically use all cores in the machine, unless
        otherwise specified in number_of_processors.

        Data with multiple TE are normalized in separate segments using the
        b0-values according that TE.

        Parameters
        ----------
        acquisition_scheme : DmipyAcquisitionScheme instance,
            An acquisition scheme that has been instantiated using dMipy.
        data : N-dimensional array of size (N_x, N_y, ..., N_dwis),
            The measured DWI signal attenuation array of either a single voxel
            or an N-dimensional dataset.
        mask : (N-1)-dimensional integer/boolean array of size (N_x, N_y, ...),
            Optional mask of voxels to be included in the optimization.
        solver : string,
            Selection of optimization algorithm.
            - 'brute2fine' to use brute-force optimization.
            - 'mix' to use Microstructure Imaging of Crossing (MIX)
              optimization.
        Ns : integer,
            for brute optimization, decised how many steps are sampled for
            every parameter.
        maxiter : integer,
            for MIX optimization, how many iterations are allowed.
        N_sphere_samples : integer,
            for brute optimization, how many spherical orientations are sampled
            for 'mu'.
        use_parallel_processing : bool,
            whether or not to use parallel processing using pathos.
        number_of_processors : integer,
            number of processors to use for parallel processing. Defaults to
            the number of processors in the computer according to cpu_count().

        Returns
        -------
        FittedCompartmentModel: class instance that contains fitted parameters,
            Can be used to recover parameters themselves or other useful
            functions.
        """
        self._check_tissue_model_acquisition_scheme(acquisition_scheme)
        self._check_model_params_with_acquisition_params(acquisition_scheme)
        self._check_acquisition_scheme_has_b0s(acquisition_scheme)
        self._check_if_volume_fractions_are_fixed()
        self._check_if_sh_coeff_fixed_if_present()

        # estimate S0
        self.scheme = acquisition_scheme
        data_ = np.atleast_2d(data)
        if self.scheme.TE is None or len(np.unique(self.scheme.TE)) == 1:
            S0 = np.mean(data_[..., self.scheme.b0_mask], axis=-1)
        else:  # if multiple TE are in the data
            S0 = np.ones_like(data_)
            for TE_ in self.scheme.shell_TE:
                TE_mask = self.scheme.TE == TE_
                TE_b0_mask = np.all([self.scheme.b0_mask, TE_mask], axis=0)
                S0[..., TE_mask] = np.mean(
                    data_[..., TE_b0_mask], axis=-1)[..., None]

        if mask is None:
            mask = data_[..., 0] > 0
        else:
            mask = np.all([mask, data_[..., 0] > 0], axis=0)
        mask_pos = np.where(mask)

        N_parameters = len(self.bounds_for_optimization)
        N_voxels = np.sum(mask)

        # make starting parameters and data the same size
        x0_ = self.parameter_initial_guess_to_parameter_vector(
            **self.x0_parameters)
        x0_ = homogenize_x0_to_data(
            data_, x0_)
        x0_bool = np.all(
            np.isnan(x0_), axis=tuple(np.arange(x0_.ndim - 1)))
        x0_[..., ~x0_bool] /= self.scales_for_optimization[~x0_bool]

        if use_parallel_processing and not have_pathos:
            msg = 'Cannot use parallel processing without pathos.'
            raise ValueError(msg)
        elif use_parallel_processing and have_pathos:
            fitted_parameters_lin = [None] * N_voxels
            if number_of_processors is None:
                number_of_processors = cpu_count()
            pool = pp.ProcessPool(number_of_processors)
            print('Using parallel processing with {} workers.'.format(
                number_of_processors))
        else:
            fitted_parameters_lin = np.empty(
                np.r_[N_voxels, N_parameters], dtype=float)

        start = time()
        if solver == 'brute2fine':
            global_brute = GlobalBruteOptimizer(
                self, self.scheme, x0_, Ns, N_sphere_samples)
            fit_func = Brute2FineOptimizer(self, self.scheme, Ns)
            print('Setup brute2fine optimizer in {} seconds'.format(
                time() - start))
        elif solver == 'mix':
            self._check_for_tortuosity_constraint()
            fit_func = MixOptimizer(self, self.scheme, maxiter)
            print('Setup MIX optimizer in {} seconds'.format(
                time() - start))
        else:
            msg = "Unknown solver name {}".format(solver)
            raise ValueError(msg)
        self.optimizer = fit_func

        start = time()
        for idx, pos in enumerate(zip(*mask_pos)):
            voxel_E = data_[pos] / S0[pos]
            voxel_x0_vector = x0_[pos]
            if solver == 'brute2fine':
                if global_brute.global_optimization_grid is True:
                    voxel_x0_vector = global_brute(voxel_E)
            fit_args = (voxel_E, voxel_x0_vector)

            if use_parallel_processing:
                fitted_parameters_lin[idx] = pool.apipe(fit_func, *fit_args)
            else:
                fitted_parameters_lin[idx] = fit_func(*fit_args)
        if use_parallel_processing:
            fitted_parameters_lin = np.array(
                [p.get() for p in fitted_parameters_lin])
            pool.close()
            pool.join()
            pool.clear()

        fitting_time = time() - start
        print('Fitting of {} voxels complete in {} seconds.'.format(
            len(fitted_parameters_lin), fitting_time))
        print('Average of {} seconds per voxel.'.format(
            fitting_time / N_voxels))

        fitted_mt_fractions = None
        if self.S0_tissue_responses:
            # secondary fitting including S0 responses
            print('Starting secondary multi-tissue optimization.')
            start = time()
            mt_fractions = np.empty(
                np.r_[N_voxels, self.N_models], dtype=float)
            fit_func = MultiTissueConvexOptimizer(
                acquisition_scheme, self, self.S0_tissue_responses)
            for idx, pos in enumerate(zip(*mask_pos)):
                voxel_S = data_[pos]
                parameters = fitted_parameters_lin[idx]
                mt_fractions[idx] = fit_func(voxel_S, parameters)
            fitting_time = time() - start
            msg = 'Multi-tissue fitting of {} voxels complete in {} seconds.'
            print(msg.format(len(mt_fractions), fitting_time))
            fitted_mt_fractions = np.zeros(np.r_[mask.shape, self.N_models])
            fitted_mt_fractions[mask_pos] = mt_fractions

        fitted_parameters = np.zeros_like(x0_, dtype=float)
        fitted_parameters[mask_pos] = (
            fitted_parameters_lin * self.scales_for_optimization)

        return FittedMultiCompartmentModel(self, S0, mask, fitted_parameters,
                                           fitted_mt_fractions)

    def simulate_signal(self, acquisition_scheme, parameters_array_or_dict):
        """
        Function to simulate diffusion data for a given acquisition_scheme
        and model parameters for the MultiCompartmentModel.

        Parameters
        ----------
        acquisition_scheme : DmipyAcquisitionScheme instance,
            An acquisition scheme that has been instantiated using dMipy
        model_parameters_array : 1D array of size (N_parameters) or
            N-dimensional array the same size as the data.
            The model parameters of the MultiCompartmentModel model.

        Returns
        -------
        E_simulated: 1D array of size (N_parameters) or N-dimensional
            array the same size as x0.
            The simulated signal of the microstructure model.
        """
        self._check_model_params_with_acquisition_params(acquisition_scheme)

        Ndata = acquisition_scheme.number_of_measurements
        if isinstance(parameters_array_or_dict, np.ndarray):
            x0 = parameters_array_or_dict
        elif isinstance(parameters_array_or_dict, dict):
            x0 = self.parameters_to_parameter_vector(
                **parameters_array_or_dict)

        x0_at_least_2d = np.atleast_2d(x0)
        x0_2d = x0_at_least_2d.reshape(-1, x0_at_least_2d.shape[-1])
        E_2d = np.empty(np.r_[x0_2d.shape[:-1], Ndata])
        for i, x0_ in enumerate(x0_2d):
            parameters = self.parameter_vector_to_parameters(x0_)
            E_2d[i] = self(acquisition_scheme, **parameters)
        E_simulated = E_2d.reshape(
            np.r_[x0_at_least_2d.shape[:-1], Ndata])

        if x0.ndim == 1:
            return np.squeeze(E_simulated)
        else:
            return E_simulated

    def __call__(self, acquisition_scheme_or_vertices,
                 quantity="signal", **kwargs):
        """
        The MultiCompartmentModel function call for to generate signal
        attenuation for a given acquisition scheme and model parameters.

        First, the linked parameters are added to the optimized parameters.

        Then, every model in the MultiCompartmentModel is called with the right
        parameters to recover the part of the signal attenuation of that model.
        The resulting values are multiplied with the volume fractions and
        finally the combined signal attenuation is returned.

        Aside from the signal, the function call can also return the Fiber
        Orientation Distributions (FODs) when a dispersed model is used, and
        can also return the stochastic cost function for the MIX algorithm.

        Parameters
        ----------
        acquisition_scheme : DmipyAcquisitionScheme instance,
            An acquisition scheme that has been instantiated using dMipy.
        quantity : string
            can be 'signal', 'FOD' or 'stochastic cost function' depending on
            the need of the model.
        kwargs: keyword arguments to the model parameter values,
            Is internally given as **parameter_dictionary.
        """
        if quantity == "signal" or quantity == "FOD":
            values = 0
        elif quantity == "stochastic cost function":
            values = np.empty((
                acquisition_scheme_or_vertices.number_of_measurements,
                len(self.models)
            ))
            counter = 0

        kwargs = self.add_linked_parameters_to_parameters(
            kwargs
        )
        if len(self.models) > 1:
            partial_volumes = [
                kwargs[p] for p in self.partial_volume_names
            ]
        else:
            partial_volumes = [1.]

        for model_name, model, partial_volume in zip(
                self.model_names, self.models, partial_volumes
        ):
            parameters = {}
            for parameter in model.parameter_ranges:
                parameter_name = self._inverted_parameter_map[
                    (model, parameter)
                ]
                parameters[parameter] = kwargs.get(
                    # , self.parameter_defaults.get(parameter_name)
                    parameter_name
                )

            if quantity == "signal":
                values = (values + partial_volume * model(
                    acquisition_scheme_or_vertices, **parameters))
            elif quantity == "FOD":
                try:
                    values = (values + partial_volume * model.fod(
                        acquisition_scheme_or_vertices, **parameters))
                except AttributeError:
                    continue
            elif quantity == "stochastic cost function":
                values[:, counter] = model(acquisition_scheme_or_vertices,
                                           **parameters)
                counter += 1
        return values


class MultiCompartmentSphericalMeanModel(MultiCompartmentModelProperties):
    r'''
    The MultiCompartmentModel class allows to combine any number of
    CompartmentModels and DistributedModels into one combined model that can
    be used to fit and simulate dMRI data.

    Parameters
    ----------
    models : list of N CompartmentModel instances,
        the models to combine into the MultiCompartmentModel.
    parameter_links : list of iterables (model, parameter name, link function,
        argument list),
        deprecated, for testing only.
    '''

    def __init__(self, models, S0_tissue_responses=None, parameter_links=None):
        self.models = models
        self.N_models = len(models)
        if S0_tissue_responses is not None:
            if len(S0_tissue_responses) != self.N_models:
                msg = 'Number of S0_tissue responses {} must be same as ' \
                      'number of input models {}.'
                raise ValueError(
                    msg.format(len(S0_tissue_responses), self.N_models))
        self.S0_tissue_responses = S0_tissue_responses
        self.parameter_links = parameter_links
        if parameter_links is None:
            self.parameter_links = []

        self._check_for_NMR_models()
        self._prepare_parameters()
        self._delete_orientation_parameters()
        self._prepare_partial_volumes()
        self._prepare_parameter_links()
        self._prepare_model_properties()
        self._check_for_double_model_class_instances()
        self._prepare_parameters_to_optimize()
        self.x0_parameters = {}

        if not have_numba:
            msg = "We highly recommend installing numba for faster function "
            msg += "execution and model fitting."
            print(msg)
        if not have_pathos:
            msg = "We highly recommend installing pathos to take advantage of "
            msg += "multicore processing."
            print(msg)

    def _check_for_NMR_models(self):
        for model in self.models:
            if model._model_type == 'NMRModel':
                msg = "Cannot estimate spherical mean of 1D-NMR models."
                raise ValueError(msg)

    def _delete_orientation_parameters(self):
        """
        Deletes orientation parameters from input models 'mu' since they're not
        needed in spherical mean models.
        """
        "Removes orientation parameters from input models."
        for model in self.models:
            for param_name, param_type in model.parameter_types.items():
                if param_type == 'orientation':
                    appended_param_name = self._inverted_parameter_map[
                        model, param_name]
                    del self.parameter_ranges[appended_param_name]
                    del self.parameter_scales[appended_param_name]
                    del self.parameter_cardinality[appended_param_name]
                    del self.parameter_types[appended_param_name]

    def fit(self, acquisition_scheme, data,
            mask=None, solver='brute2fine', Ns=5, maxiter=300,
            N_sphere_samples=30, use_parallel_processing=have_pathos,
            number_of_processors=None):
        """ The main data fitting function of a MultiCompartmentModel.

        This function can fit it to an N-dimensional dMRI data set, and returns
        a FittedMultiCompartmentModel instance that contains the fitted
        parameters and other useful functions to study the results.

        No initial guess needs to be given to fit a model, but a partial or
        complete initial guess can be given if the user wants to have a
        solution that is a local minimum close to that guess. The
        parameter_initial_guess input can be created using
        parameter_initial_guess_to_parameter_vector().

        A mask can also be given to exclude voxels from fitting (e.g. voxels
        that are outside the brain). If no mask is given then all voxels are
        included.

        An optimization approach can be chosen as either 'brute2fine' or 'mix'.
        - Choosing brute2fine will first use a brute-force optimization to find
          an initial guess for parameters without one, and will then refine the
          result using gradient-descent-based optimization.

          Note that given no initial guess will make brute2fine precompute an
          global parameter grid that will be re-used for all voxels, which in
          many cases is much faster than giving voxel-varying initial condition
          that requires a grid to be estimated per voxel.

        - Choosing mix will use the recent MIX algorithm based on separation of
          linear and non-linear parameters. MIX first uses a stochastic
          algorithm to find the non-linear parameters (non-volume fractions),
          then estimates the volume fractions while fixing the estimates of the
          non-linear parameters, and then finally refines the solution using
          a gradient-descent-based algorithm.

        The fitting process can be readily parallelized using the optional
        "pathos" package. If it is installed then it will automatically use it,
        but it can be turned off by setting use_parallel_processing=False. The
        algorithm will automatically use all cores in the machine, unless
        otherwise specified in number_of_processors.

        Data with multiple TE are normalized in separate segments using the
        b0-values according that TE.

        Parameters
        ----------
        acquisition_scheme : DmipyAcquisitionScheme instance,
            An acquisition scheme that has been instantiated using dMipy.
        data : N-dimensional array of size (N_x, N_y, ..., N_dwis),
            The measured DWI signal attenuation array of either a single voxel
            or an N-dimensional dataset.
        mask : (N-1)-dimensional integer/boolean array of size (N_x, N_y, ...),
            Optional mask of voxels to be included in the optimization.
        solver : string,
            Selection of optimization algorithm.
            - 'brute2fine' to use brute-force optimization.
            - 'mix' to use Microstructure Imaging of Crossing (MIX)
              optimization.
        Ns : integer,
            for brute optimization, decised how many steps are sampled for
            every parameter.
        maxiter : integer,
            for MIX optimization, how many iterations are allowed.
        N_sphere_samples : integer,
            for brute optimization, how many spherical orientations are sampled
            for 'mu'.
        use_parallel_processing : bool,
            whether or not to use parallel processing using pathos.
        number_of_processors : integer,
            number of processors to use for parallel processing. Defaults to
            the number of processors in the computer according to cpu_count().

        Returns
        -------
        FittedCompartmentModel: class instance that contains fitted parameters,
            Can be used to recover parameters themselves or other useful
            functions.
        """
        self._check_tissue_model_acquisition_scheme(acquisition_scheme)
        self._check_model_params_with_acquisition_params(acquisition_scheme)
        self._check_acquisition_scheme_has_b0s(acquisition_scheme)
        self._check_if_volume_fractions_are_fixed()

        # estimate S0
        self.scheme = acquisition_scheme
        data_ = np.atleast_2d(data)
        if self.scheme.TE is None or len(np.unique(self.scheme.TE)) == 1:
            S0 = np.mean(data_[..., self.scheme.b0_mask], axis=-1)
        else:  # if multiple TE are in the data
            S0 = np.ones(np.r_[data_.shape[:-1],
                               len(acquisition_scheme.shell_TE)])
            for TE_ in self.scheme.shell_TE:
                TE_mask = self.scheme.shell_TE == TE_
                TE_mask_shell = self.scheme.TE == TE_
                TE_b0_mask = np.all([self.scheme.b0_mask, TE_mask_shell],
                                    axis=0)
                S0[..., TE_mask] = np.mean(
                    data_[..., TE_b0_mask], axis=-1)[..., None]

        if mask is None:
            mask = data_[..., 0] > 0
        else:
            mask = np.all([mask, data_[..., 0] > 0], axis=0)
        mask_pos = np.where(mask)

        N_parameters = len(self.bounds_for_optimization)
        N_voxels = np.sum(mask)

        # make starting parameters and data the same size
        x0_ = self.parameter_initial_guess_to_parameter_vector(
            **self.x0_parameters)
        x0_ = homogenize_x0_to_data(
            data_, x0_)
        x0_bool = np.all(
            np.isnan(x0_), axis=tuple(np.arange(x0_.ndim - 1)))
        x0_[..., ~x0_bool] /= self.scales_for_optimization[~x0_bool]

        if use_parallel_processing and not have_pathos:
            msg = 'Cannot use parallel processing without pathos.'
            raise ValueError(msg)
        elif use_parallel_processing and have_pathos:
            fitted_parameters_lin = [None] * N_voxels
            if number_of_processors is None:
                number_of_processors = cpu_count()
            pool = pp.ProcessPool(number_of_processors)
            print('Using parallel processing with {} workers.'.format(
                number_of_processors))
        else:
            fitted_parameters_lin = np.empty(
                np.r_[N_voxels, N_parameters], dtype=float)

        # estimate the spherical mean of the data.
        data_to_fit = np.zeros(np.r_[data_.shape[:-1], self.scheme.N_shells])
        for pos in zip(*mask_pos):
            data_to_fit[pos] = estimate_spherical_mean_multi_shell(
                data_[pos], self.scheme)

        start = time()
        if solver == 'brute2fine':
            global_brute = GlobalBruteOptimizer(
                self, self.scheme,
                x0_, Ns, N_sphere_samples)
            fit_func = Brute2FineOptimizer(self, self.scheme, Ns)
            print('Setup brute2fine optimizer in {} seconds'.format(
                time() - start))
        elif solver == 'mix':
            self._check_for_tortuosity_constraint()
            fit_func = MixOptimizer(self, self.scheme, maxiter)
            print('Setup MIX optimizer in {} seconds'.format(
                time() - start))
        else:
            msg = "Unknown solver name {}".format(solver)
            raise ValueError(msg)
        self.optimizer = fit_func

        start = time()
        for idx, pos in enumerate(zip(*mask_pos)):
            voxel_E = data_to_fit[pos] / S0[pos]
            voxel_x0_vector = x0_[pos]
            if solver == 'brute2fine':
                if global_brute.global_optimization_grid is True:
                    voxel_x0_vector = global_brute(voxel_E)
            fit_args = (voxel_E, voxel_x0_vector)

            if use_parallel_processing:
                fitted_parameters_lin[idx] = pool.apipe(fit_func, *fit_args)
            else:
                fitted_parameters_lin[idx] = fit_func(*fit_args)
        if use_parallel_processing:
            fitted_parameters_lin = np.array(
                [p.get() for p in fitted_parameters_lin])
            pool.close()
            pool.join()
            pool.clear()

        fitting_time = time() - start
        print('Fitting of {} voxels complete in {} seconds.'.format(
            len(fitted_parameters_lin), fitting_time))
        print('Average of {} seconds per voxel.'.format(
            fitting_time / N_voxels))

        fitted_mt_fractions = None
        if self.S0_tissue_responses:
            # secondary fitting including S0 responses
            print('Starting secondary multi-tissue optimization.')
            start = time()
            mt_fractions = np.empty(
                np.r_[N_voxels, self.N_models], dtype=float)
            fit_func = MultiTissueConvexOptimizer(
                acquisition_scheme, self, self.S0_tissue_responses)
            for idx, pos in enumerate(zip(*mask_pos)):
                voxel_S = data_to_fit[pos]
                parameters = fitted_parameters_lin[idx]
                mt_fractions[idx] = fit_func(voxel_S, parameters)
            fitting_time = time() - start
            msg = 'Multi-tissue fitting of {} voxels complete in {} seconds.'
            print(msg.format(len(mt_fractions), fitting_time))
            fitted_mt_fractions = np.zeros(np.r_[mask.shape, self.N_models])
            fitted_mt_fractions[mask_pos] = mt_fractions

        fitted_parameters = np.zeros_like(x0_, dtype=float)
        fitted_parameters[mask_pos] = (
            fitted_parameters_lin * self.scales_for_optimization)

        return FittedMultiCompartmentSphericalMeanModel(
            self, S0, mask, fitted_parameters, fitted_mt_fractions)

    def simulate_signal(self, acquisition_scheme, parameters_array_or_dict):
        """
        Function to simulate diffusion data for a given acquisition_scheme
        and model parameters for the MultiCompartmentModel.

        Parameters
        ----------
        acquisition_scheme : DmipyAcquisitionScheme instance,
            An acquisition scheme that has been instantiated using dMipy
        model_parameters_array : 1D array of size (N_parameters) or
            N-dimensional array the same size as the data.
            The model parameters of the MultiCompartmentModel model.

        Returns
        -------
        E_simulated: 1D array of size (N_parameters) or N-dimensional
            array the same size as x0.
            The simulated signal of the microstructure model.
        """
        self._check_model_params_with_acquisition_params(acquisition_scheme)

        Ndata = acquisition_scheme.shell_indices.max() + 1
        if isinstance(parameters_array_or_dict, np.ndarray):
            x0 = parameters_array_or_dict
        elif isinstance(parameters_array_or_dict, dict):
            x0 = self.parameters_to_parameter_vector(
                **parameters_array_or_dict)

        x0_at_least_2d = np.atleast_2d(x0)
        x0_2d = x0_at_least_2d.reshape(-1, x0_at_least_2d.shape[-1])
        E_2d = np.empty(np.r_[x0_2d.shape[:-1], Ndata])
        for i, x0_ in enumerate(x0_2d):
            parameters = self.parameter_vector_to_parameters(x0_)
            E_2d[i] = self(acquisition_scheme, **parameters)
        E_simulated = E_2d.reshape(
            np.r_[x0_at_least_2d.shape[:-1], Ndata])

        if x0.ndim == 1:
            return np.squeeze(E_simulated)
        else:
            return E_simulated

    def __call__(self, acquisition_scheme_or_vertices,
                 quantity="signal", **kwargs):
        """
        The MultiCompartmentModel function call for to generate signal
        attenuation for a given acquisition scheme and model parameters.

        First, the linked parameters are added to the optimized parameters.

        Then, every model in the MultiCompartmentModel is called with the right
        parameters to recover the part of the signal attenuation of that model.
        The resulting values are multiplied with the volume fractions and
        finally the combined signal attenuation is returned.

        Aside from the signal, the function call can also return the Fiber
        Orientation Distributions (FODs) when a dispersed model is used, and
        can also return the stochastic cost function for the MIX algorithm.

        Parameters
        ----------
        acquisition_scheme : DmipyAcquisitionScheme instance,
            An acquisition scheme that has been instantiated using dMipy.
        quantity : string
            can be 'signal', 'FOD' or 'stochastic cost function' depending on
            the need of the model.
        kwargs: keyword arguments to the model parameter values,
            Is internally given as **parameter_dictionary.
        """
        if quantity == "signal":
            values = 0
        elif quantity == "stochastic cost function":
            values = np.empty((
                len(acquisition_scheme_or_vertices.shell_bvalues),
                len(self.models)
            ))
            counter = 0

        kwargs = self.add_linked_parameters_to_parameters(
            kwargs
        )
        if len(self.models) > 1:
            partial_volumes = [
                kwargs[p] for p in self.partial_volume_names
            ]
        else:
            partial_volumes = [1.]

        for model_name, model, partial_volume in zip(
                self.model_names, self.models, partial_volumes
        ):
            parameters = {}
            for parameter in model.parameter_ranges:
                parameter_name = self._inverted_parameter_map[
                    (model, parameter)
                ]
                parameters[parameter] = kwargs.get(
                    # , self.parameter_defaults.get(parameter_name)
                    parameter_name
                )

            if quantity == "signal":
                values = (values + partial_volume * model.spherical_mean(
                    acquisition_scheme_or_vertices, **parameters))
            elif quantity == "stochastic cost function":
                values[:, counter] = model.spherical_mean(
                    acquisition_scheme_or_vertices,
                    **parameters)
                counter += 1
        return values


class MultiCompartmentSphericalHarmonicsModel(MultiCompartmentModelProperties):
    r'''
    The MultiCompartmentModel class allows to combine any number of
    CompartmentModels and DistributedModels into one combined model that can
    be used to fit and simulate dMRI data.

    Parameters
    ----------
    models : list of N CompartmentModel instances,
        the models to combine into the MultiCompartmentModel.
    '''

    def __init__(self, models, S0_tissue_responses=None, sh_order=8):
        self.models = models
        self.N_models = len(models)
        if S0_tissue_responses is not None:
            self.fit_S0_response = True
            if len(S0_tissue_responses) != self.N_models:
                msg = 'Number of S0_tissue responses {} must be same as ' \
                      'number of input models {}.'
                raise ValueError(
                    msg.format(len(S0_tissue_responses), self.N_models))
        else:
            self.fit_S0_response = False
        self.S0_tissue_responses = S0_tissue_responses
        self.parameter_links = []

        self._check_for_NMR_models()
        self._prepare_parameters()
        self._delete_orientation_parameters()
        self._prepare_partial_volumes()
        self._prepare_parameter_links()
        self._prepare_model_properties()
        self._check_for_double_model_class_instances()
        self._prepare_parameters_to_optimize()
        self._add_spherical_harmonics_parameters(sh_order)
        self._check_that_one_anisotropic_kernel_is_present()
        # self._check_for_tissue_response_models()

        self.x0_parameters = {}
        self.sh_order = sh_order

        if not have_numba:
            msg = "We highly recommend installing numba for faster function "
            msg += "execution and model fitting."
            print(msg)
        if not have_pathos:
            msg = "We highly recommend installing pathos to take advantage of "
            msg += "multicore processing."
            print(msg)

    def _check_for_NMR_models(self):
        for model in self.models:
            if model._model_type == 'NMRModel':
                msg = "Cannot estimate spherical mean of 1D-NMR models."
                raise ValueError(msg)

    def _delete_orientation_parameters(self):
        """
        Deletes orientation parameters from input models 'mu' since they're not
        needed in spherical mean models.
        """
        "Removes orientation parameters from input models."
        for model in self.models:
            for param_name, param_type in model.parameter_types.items():
                if param_type == 'orientation':
                    appended_param_name = self._inverted_parameter_map[
                        model, param_name]
                    del self.parameter_ranges[appended_param_name]
                    del self.parameter_scales[appended_param_name]
                    del self.parameter_cardinality[appended_param_name]
                    del self.parameter_types[appended_param_name]

    def _add_spherical_harmonics_parameters(self, sh_order):
        N_coef = int((sh_order + 2) * (sh_order + 1) // 2)
        self.parameter_ranges['sh_coeff'] = [
            [-1e3, 1e3] for i in range(N_coef)]
        self.parameter_scales['sh_coeff'] = np.ones(N_coef, dtype=float)
        self.parameter_cardinality['sh_coeff'] = N_coef
        self.parameter_types['sh_coeff'] = 'sh_coefficients'
        self.parameter_optimization_flags['sh_coeff'] = True

    def _check_if_kernel_parameters_are_fixed(self):
        "checks if only volume fraction and sh_coeff parameters are optimized."
        self.volume_fractions_fixed = True
        for name, flag in self.parameter_optimization_flags.items():
            if flag is True:
                if (not name == 'sh_coeff' and
                        not name.startswith('partial_volume')):
                    msg = 'Kernel parameter {} is not fixed.'.format(name)
                    raise ValueError(msg)
                if name.startswith('partial_volume'):
                    self.volume_fractions_fixed = False
        if (not self.volume_fractions_fixed and
                self.multiple_anisotropic_kernels):
            msg = 'Cannot have multiple anisotropic kernels without having '
            msg += 'all volume fractions fixed.'
            raise ValueError(msg)

    def _check_that_one_anisotropic_kernel_is_present(self):
        "checks if one anisotropic kernel is given."
        orientation_counter = 0
        self.multiple_anisotropic_kernels = False
        for model in self.models:
            if 'orientation' in model.parameter_types.values():
                orientation_counter += 1
        if orientation_counter == 0:
            msg = 'MultiCompartmentSphericalHarmonicsModel must at least have '
            msg += 'one anisotropic kernel input model.'
            raise ValueError(msg)
        if orientation_counter > 1:
            self.multiple_anisotropic_kernels = True

    def fit(self, acquisition_scheme, data, mask=None, solver='csd',
            lambda_lb=1e-5, unity_constraint='kernel_dependent',
            use_parallel_processing=have_pathos,
            number_of_processors=None, verbose=True):
        """ The main data fitting function of a
        MultiCompartmentSphericalHarmonicsModel.

        This function can fit it to an N-dimensional dMRI data set, and returns
        a FittedMultiCompartmentModel instance that contains the fitted
        parameters and other useful functions to study the results.

        A mask can also be given to exclude voxels from fitting (e.g. voxels
        that are outside the brain). If no mask is given then all voxels are
        included.

        The fitting process can be readily parallelized using the optional
        "pathos" package. If it is installed then it will automatically use it,
        but it can be turned off by setting use_parallel_processing=False. The
        algorithm will automatically use all cores in the machine, unless
        otherwise specified in number_of_processors.

        Data with multiple TE are normalized in separate segments using the
        b0-values according that TE.

        Parameters
        ----------
        acquisition_scheme : DmipyAcquisitionScheme instance,
            An acquisition scheme that has been instantiated using dMipy.
        data : N-dimensional array of size (N_x, N_y, ..., N_dwis),
            The measured DWI signal attenuation array of either a single voxel
            or an N-dimensional dataset.
        mask : (N-1)-dimensional integer/boolean array of size (N_x, N_y, ...),
            Optional mask of voxels to be included in the optimization.
        solver : string,
            Can be 'csd', 'csd_tounier07' or 'csd_cvxpy', with the default
            being 'csd'. Using 'csd' will make the algorithm automatically
            use the 'tournier07' solver [1]_ if there are no volume fractions
            to fit or they are fixed. Otherwise, the slower but more general
            cvxpy solver [2]_ is used, which follows the formulation of [3]_.
        lambda_lb: positive float,
            Weight for Laplace-Beltrami regularization to impose smoothness
            into estimated FODs, follows [4]_.
        unity_constraint: String or bool,
            Whether or not to constrain the volume fractions of the FOD to
            unity. The default is set to 'kernel_dependent', meaning it will
            enforce unity if the kernel is voxel-varying or when volume
            fractions are estimated. Otherwise unity_constraint is set to
            False.
        use_parallel_processing : bool,
            Whether or not to use parallel processing using pathos.
        number_of_processors : integer,
            Number of processors to use for parallel processing. Defaults to
            the number of processors in the computer according to cpu_count().

        Returns
        -------
        FittedCompartmentModel: class instance that contains fitted parameters,
            Can be used to recover parameters themselves or other useful
            functions.

        References
        ----------
        .. [1] Tournier, J-Donald, Fernando Calamante, and Alan Connelly.
            "Robust determination of the fibre orientation distribution in
            diffusion MRI: non-negativity constrained super-resolved spherical
            deconvolution." Neuroimage 35.4 (2007): 1459-1472.
        .. [2] Diamond, Steven, and Stephen Boyd. "CVXPY: A Python-embedded
            modeling language for convex optimization." The Journal of Machine
            Learning Research 17.1 (2016): 2909-2913.
        .. [3] Jeurissen, Ben, et al. "Multi-tissue constrained spherical
            deconvolution for improved analysis of multi-shell diffusion MRI
            data." NeuroImage 103 (2014): 411-426.
        .. [4] Descoteaux, Maxime, et al. "Regularized, fast, and robust
            analytical Q-ball imaging." Magnetic Resonance in Medicine: An
            Official Journal of the International Society for Magnetic
            Resonance in Medicine 58.3 (2007): 497-510.
        """
        self._check_if_kernel_parameters_are_fixed()
        self._check_tissue_model_acquisition_scheme(acquisition_scheme)
        self._check_acquisition_scheme_has_b0s(acquisition_scheme)
        self._check_model_params_with_acquisition_params(acquisition_scheme)

        self.voxel_varying_kernel = False
        if bool(self.x0_parameters):  # if the dictionary is not empty
            self.voxel_varying_kernel = True

        if unity_constraint == 'kernel_dependent':
            self.unity_constraint = False
            if self.fit_S0_response:
                self.unity_constraint = False
            elif not self.volume_fractions_fixed or self.voxel_varying_kernel:
                self.unity_constraint = True
        else:
            self.unity_constraint = unity_constraint

        if self.fit_S0_response:
            S0_responses = np.array(self.S0_tissue_responses)
            self.max_S0_response = S0_responses.max()
            self.S0_responses = S0_responses / self.max_S0_response
        else:
            self.S0_responses = np.ones(len(self.models), dtype=float)
            self.max_S0_response = 1.

        # estimate S0
        self.scheme = acquisition_scheme
        data_ = np.atleast_2d(data)
        if self.scheme.TE is None or len(np.unique(self.scheme.TE)) == 1:
            S0 = np.mean(data_[..., self.scheme.b0_mask], axis=-1)
        else:  # if multiple TE are in the data
            S0 = np.ones_like(data_)
            for TE_ in self.scheme.shell_TE:
                TE_mask = self.scheme.TE == TE_
                TE_b0_mask = np.all([self.scheme.b0_mask, TE_mask], axis=0)
                S0[..., TE_mask] = np.mean(
                    data_[..., TE_b0_mask], axis=-1)[..., None]

        if mask is None:
            mask = data_[..., 0] > 0
        else:
            mask = np.all([mask, data_[..., 0] > 0], axis=0)
        mask_pos = np.where(mask)

        N_parameters = len(self.bounds_for_optimization)
        N_voxels = np.sum(mask)

        # make starting parameters and data the same size
        x0_ = self.parameter_initial_guess_to_parameter_vector(
            **self.x0_parameters)
        x0_ = homogenize_x0_to_data(
            data_, x0_)

        start = time()
        if solver == 'csd':
            if self.volume_fractions_fixed:
                fit_func = CsdTournierOptimizer(
                    acquisition_scheme, self, x0_, self.sh_order,
                    unity_constraint=self.unity_constraint,
                    lambda_lb=lambda_lb)
                if use_parallel_processing:
                    msg = 'Parallel processing turned off for tournier07'
                    msg += ' optimizer because it does not improve fitting '
                    msg += 'speed.'
                    if verbose:
                        print(msg)
                    use_parallel_processing = False
                if verbose:
                    print(
                        'Setup Tournier07 FOD optimizer in {} seconds'.format(
                            time() - start))
            else:
                fit_func = CsdCvxpyOptimizer(
                    acquisition_scheme, self, x0_, self.sh_order,
                    unity_constraint=self.unity_constraint,
                    lambda_lb=lambda_lb)
                if verbose:
                    print('Setup CVXPY FOD optimizer in {} seconds'.format(
                        time() - start))
        elif solver == 'csd_tournier07':
            fit_func = CsdTournierOptimizer(
                acquisition_scheme, self, x0_, self.sh_order,
                unity_constraint=self.unity_constraint, lambda_lb=lambda_lb)
            if use_parallel_processing:
                msg = 'Parallel processing turned off for tournier07 optimizer'
                msg += ' because it does not improve fitting speed.'
                if verbose:
                    print(msg)
                use_parallel_processing = False
            if verbose:
                print('Setup Tournier07 FOD optimizer in {} seconds'.format(
                    time() - start))
        elif solver == 'csd_cvxpy':
            fit_func = CsdCvxpyOptimizer(
                acquisition_scheme, self, x0_, self.sh_order,
                unity_constraint=self.unity_constraint, lambda_lb=lambda_lb)
            if verbose:
                print('Setup CVXPY FOD optimizer in {} seconds'.format(
                    time() - start))
        else:
            msg = "Unknown solver name {}".format(solver)
            raise ValueError(msg)

        self.optimizer = fit_func

        if use_parallel_processing and not have_pathos:
            msg = 'Cannot use parallel processing without pathos.'
            raise ValueError(msg)
        elif use_parallel_processing and have_pathos:
            fitted_parameters_lin = [None] * N_voxels
            if number_of_processors is None:
                number_of_processors = cpu_count()
            pool = pp.ProcessPool(number_of_processors)
            if verbose:
                print('Using parallel processing with {} workers.'.format(
                    number_of_processors))
        else:
            fitted_parameters_lin = np.empty(
                np.r_[N_voxels, N_parameters], dtype=float)

        start = time()
        for idx, pos in enumerate(zip(*mask_pos)):
            if self.fit_S0_response:
                data_to_fit = data_[pos] / self.max_S0_response
            else:
                data_to_fit = data_[pos] / S0[pos]
            voxel_x0_vector = x0_[pos]
            fit_args = (data_to_fit, voxel_x0_vector)

            if use_parallel_processing:
                fitted_parameters_lin[idx] = pool.apipe(fit_func, *fit_args)
            else:
                fitted_parameters_lin[idx] = fit_func(*fit_args)
        if use_parallel_processing:
            fitted_parameters_lin = np.array(
                [p.get() for p in fitted_parameters_lin])
            pool.close()
            pool.join()
            pool.clear()

        fitting_time = time() - start
        if verbose:
            print('Fitting of {} voxels complete in {} seconds.'.format(
                len(fitted_parameters_lin), fitting_time))
            print('Average of {} seconds per voxel.'.format(
                fitting_time / N_voxels))
        fitted_parameters = np.zeros_like(x0_, dtype=float)
        fitted_parameters[mask_pos] = fitted_parameters_lin

        return FittedMultiCompartmentSphericalHarmonicsModel(
            self, S0, mask, fitted_parameters)

    def simulate_signal(self, acquisition_scheme, parameters_array_or_dict):
        """
        Function to simulate diffusion data for a given acquisition_scheme
        and model parameters for the MultiCompartmentModel.

        Parameters
        ----------
        acquisition_scheme : DmipyAcquisitionScheme instance,
            An acquisition scheme that has been instantiated using dMipy
        model_parameters_array : 1D array of size (N_parameters) or
            N-dimensional array the same size as the data.
            The model parameters of the MultiCompartmentModel model.

        Returns
        -------
        E_simulated: 1D array of size (N_parameters) or N-dimensional
            array the same size as x0.
            The simulated signal of the microstructure model.
        """
        self._check_model_params_with_acquisition_params(acquisition_scheme)
        self.volume_fractions_fixed = True
        if self.S0_tissue_responses is None:
            self.S0_responses = np.ones(self.N_models)

        Ndata = acquisition_scheme.number_of_measurements
        if isinstance(parameters_array_or_dict, np.ndarray):
            x0 = parameters_array_or_dict
        elif isinstance(parameters_array_or_dict, dict):
            x0 = self.parameters_to_parameter_vector(
                **parameters_array_or_dict)

        x0_at_least_2d = np.atleast_2d(x0)
        x0_2d = x0_at_least_2d.reshape(-1, x0_at_least_2d.shape[-1])
        E_2d = np.empty(np.r_[x0_2d.shape[:-1], Ndata])
        for i, x0_ in enumerate(x0_2d):
            parameters = self.parameter_vector_to_parameters(x0_)
            E_2d[i] = self(acquisition_scheme, **parameters)
        E_simulated = E_2d.reshape(
            np.r_[x0_at_least_2d.shape[:-1], Ndata])

        if x0.ndim == 1:
            return np.squeeze(E_simulated)
        else:
            return E_simulated

    def __call__(self, acquisition_scheme, **kwargs):
        """
        The MultiCompartmentModel function call for to generate signal
        attenuation for a given acquisition scheme and model parameters.

        First, the linked parameters are added to the optimized parameters.

        Then, every model in the MultiCompartmentModel is called with the right
        parameters to recover the part of the signal attenuation of that model.
        The resulting values are multiplied with the volume fractions and
        finally the combined signal attenuation is returned.

        Aside from the signal, the function call can also return the Fiber
        Orientation Distributions (FODs) when a dispersed model is used, and
        can also return the stochastic cost function for the MIX algorithm.

        Parameters
        ----------
        acquisition_scheme : DmipyAcquisitionScheme instance,
            An acquisition scheme that has been instantiated using dMipy.
        kwargs: keyword arguments to the model parameter values,
            Is internally given as **parameter_dictionary.
        """
        A = self._construct_convolution_kernel(
            acquisition_scheme=acquisition_scheme, **kwargs)
        self.S0_responses = kwargs.get('S0_responses', self.S0_responses)

        # if vf fixed then just multiply with sh_coeff
        if self.volume_fractions_fixed:
            E = np.dot(A, kwargs['sh_coeff'])
        else:
            sh_coeff = np.zeros(self.optimizer.Ncoef_total)
            sh_coeff[self.optimizer.sh_start:
                     self.optimizer.Ncoef + self.optimizer.sh_start] = kwargs[
                'sh_coeff']
            for i, name in enumerate(self.partial_volume_names):
                sh_coeff[self.optimizer.vf_indices[i]] = (
                    kwargs[name] / (2 * np.sqrt(np.pi)))
            E = np.dot(A, sh_coeff)
        return E


def homogenize_x0_to_data(data, x0):
    """
    Function that checks if data and initial guess x0 are of the same size.
    If x0 is 1D, it will be tiled to be the same size as data.
    """
    if x0 is not None:
        if x0.ndim == 1:
            # the same x0 will be used for every voxel in N-dimensional data.
            x0_as_data = np.tile(x0, np.r_[data.shape[:-1], 1])
        else:
            x0_as_data = x0.copy()
    if not np.all(
            x0_as_data.shape[:-1] == data.shape[:-1]
    ):
        # if x0 and data are both N-dimensional but have different shapes.
        msg = "data and x0 both N-dimensional but have different shapes. "
        msg += "Current shapes are {} and {}.".format(
            data.shape[:-1],
            x0_as_data.shape[:-1])
        raise ValueError(msg)
    return x0_as_data


class ReturnFixedValue:
    "Parameter fixing class for parameter links."

    def __init__(self, value):
        self.value = value

    def __call__(self):
        return self.value



import numpy as np

__all__ = [
    'construct_model_based_A_matrix'
]


def construct_model_based_A_matrix(acquisition_scheme, model_rh, lmax):
    """Constructs the multi-shell observation matrix from spherical_harmonics
    to DWIs. Follows the notation of Eq. (2) in [1]_.

    The dmipy acquisition_scheme object contains all the information on which
    DWIs belong to which acquisition shells, what are the maximum spherical
    harmonics order used for each shell, and the observation matrix that maps
    the DWIs of each shell to a spherical harmonics representation.

    The dmipy model must be have all parameters fixed to be able to generate
    the rotational harmonics of each shell.

    Parameters
    ----------
    acquisition_scheme : DmipyAcquisitionScheme instance,
            An acquisition scheme that has been instantiated using dmipy.
    model_rh: array of size (N_shells, N_rh_coeffs for that shell),
        rotational harmonics for every shell.
    lmax: even positive integer,
        even maximum spherical harmonics order of the to-be-estimated FOD.

    Returns
    -------
    Ams: array of size (N_DWIs, N_sh_coef),
        observation matrix to map spherical harmonics to DWIs.

    References
    ----------
    .. [1] Jeurissen, Ben, et al. "Multi-tissue constrained spherical
        deconvolution for improved analysis of multi-shell diffusion MRI data."
        NeuroImage 103 (2014): 411-426.
    """
    Ncoef = int((lmax + 2) * (lmax + 1) // 2)
    Ams = np.zeros([acquisition_scheme.number_of_measurements, Ncoef])
    Ams[acquisition_scheme.b0_mask, 0] = 2 * np.sqrt(np.pi)

    sh_eigenvalues = np.zeros([len(model_rh), Ncoef])

    # prepare the rotational harmonics of the kernel
    counter = 0
    for n_ in range(0, lmax + 1, 2):
        if n_ // 2 >= model_rh.shape[1]:
            # in case an isotropic kernel is given
            break
        coef_in_order = 2 * n_ + 1
        sh_eigenvalues[:, counter: counter + coef_in_order] = (
            np.sqrt((4 * np.pi) / (2 * n_ + 1)) *   # sh eigenvalues
            model_rh[:, n_ // 2: n_ // 2 + 1])

        counter += coef_in_order

    # construct the multi-shell observation matrix.
    for i, shell_index in enumerate(acquisition_scheme.unique_dwi_indices):
        shell_mask = acquisition_scheme.shell_indices == shell_index
        shell_sh_matrix = acquisition_scheme.shell_sh_matrices[shell_index]
        if Ncoef < shell_sh_matrix.shape[1]:
            Ams[shell_mask, :Ncoef] = np.dot(
                shell_sh_matrix[:, :Ncoef], np.diag(sh_eigenvalues[i]))
        else:
            Ams[shell_mask, :shell_sh_matrix.shape[1]] = np.dot(
                shell_sh_matrix,
                np.diag(sh_eigenvalues[i, :shell_sh_matrix.shape[1]]))
    return Ams



import numpy as np
import cvxpy


__all__ = [
    'AmicoCvxpyOptimizer'
]


class AmicoCvxpyOptimizer:
    """
    Accelerated microstructure imaging via convex optimization (AMICO). The
    algorithm implementation is based on the work presented in [1]_, but it is
    generalized in the sense that it can perform estimation of any multi-
    compartment model parameter as well as the parameter distribution. The
    optimizer is implemented using CVXPY library [2]_.

    Limitations:
        - ...

    Parameters
    ----------
    acquisition_scheme: DmipyAcquisitionScheme instance,
        An acquisition scheme that has been instantiated using Dmipy.
    model: Dmipy MultiCompartmentModel instance,
        Contains Dmipy multi-compartment model information.
    x0_vector: Array of size (Nx),
        Vector containing probability distributions of the parameters that are
        being estimated
    lambda_1: Array of size (Nc)
        Vector containing L2 regularization weights for each compartment
    lambda_2: Array of size (Nc)
        Vector containing L1 regularization weights for each compartment
    Nt: int
        Integer equal to the number of equidistant sampling points over each
        parameter range used to create atoms of observation matrix M

    References
    ----------
    .. [1] Daducci, Alessandro, et al. "Accelerated microstructure imaging
        via convex optimization (AMICO) from diffusion MRI data."
        NeuroImage 105 (2015): 32-44.

    .. [2] Diamond, Steven, and Stephen Boyd. "CVXPY: A Python-embedded
        modeling language for convex optimization." The Journal of Machine
        Learning Research 17.1 (2016): 2909-2913.
    """

    def __init__(self, model, acquisition_scheme, x0_vector=None,
                 lambda_1=None, lambda_2=None):
        self.model = model
        self.acquisition_scheme = acquisition_scheme
        self.x0_vector = x0_vector

        if len(lambda_1) != self.model.N_models or \
                len(lambda_2) != self.model.N_models:
            raise ValueError("Number of regularization weights should"
                             "correspond to the number of compartments!")
        self.lambda_1 = lambda_1
        self.lambda_2 = lambda_2

    def __call__(self, data, M, grid, idx, x0_th=1.e-4):
        """
        The fitting function of AMICO optimizer.
        Parameters
        ----------
        data : Array of size (Ndata)
            The normalized dMRI signal attenuation to be fitted.
        M: Array of size (Ndata, Nx)
            The observation matrix containing Nx model atoms
        grid: dict
            Dictionary containing tessellation of parameters to be estimated
            for each model within multi-compartment model
        idx: dict
            Dictionary containing indices that correspond to the parameters
            to be estimated for each model within multi-compartment model
        x0_th: float
            Threshold for selecting important atoms after solving NNLS
            with L1 and L2 regularization terms

        Returns
        -------
        fitted_parameter_vector : Array of size (Np),
            Vector containing estimated parameters

        """

        # 1. Contracting matrix M and data to have one b=0 value
        M = np.vstack((np.mean(M[self.acquisition_scheme.b0_mask, :], axis=0),
                      M[~self.acquisition_scheme.b0_mask, :]))
        data = np.append(np.mean(data[self.acquisition_scheme.b0_mask]),
                         data[~self.acquisition_scheme.b0_mask])

        # 2. Selecting important atoms by solving NNLS
        # regularized with L1 and L2 norms
        x0 = cvxpy.Variable(len(self.x0_vector))

        cost = 0.5 * cvxpy.sum_squares(M * x0 - data)
        for m_idx, model_name in enumerate(self.model.model_names):
            cost += self.lambda_1[m_idx] * \
                cvxpy.norm(x0[idx[model_name]], 1)
            cost += 0.5 * self.lambda_2[m_idx] * \
                cvxpy.norm(x0[idx[model_name]], 2) ** 2
        problem = cvxpy.Problem(cvxpy.Minimize(cost), [x0 >= 0])
        problem.solve()
        self.x0_vector = x0.value

        # 3. Computing distribution vector x0_vector by solving NNLS
        x0_idx_i = self.x0_vector > x0_th
        x0_i = cvxpy.Variable(sum(x0_idx_i))
        cost = cvxpy.sum_squares(M[:, x0_idx_i] * x0_i - data)
        problem = cvxpy.Problem(cvxpy.Minimize(cost), [x0_i >= 0])
        problem.solve()
        self.x0_vector[~x0_idx_i] = 0.
        self.x0_vector[x0_idx_i] = x0_i.value
        self.x0_vector /= (np.sum(self.x0_vector) + 1.e-8)

        # 4. Estimating parameters based using estimated distribution
        # vector and tessellation grids
        fitted_parameter_vector = []
        for m_idx, model_name in enumerate(self.model.model_names):
            m = self.model.models[m_idx]
            if 'partial_volume_' + str(m_idx) in grid:
                p_estim = np.sum(self.x0_vector[idx[model_name]])
                fitted_parameter_vector.append(p_estim)
            for p in m.parameter_names:
                if p.endswith('mu'):
                    continue
                if model_name + p in grid:
                    p_estim = \
                        np.sum(grid[model_name + p][idx[model_name]] *
                               self.x0_vector[idx[model_name]]) /\
                        (np.sum(self.x0_vector[idx[model_name]]) + 1.e-8)
                    fitted_parameter_vector.append(p_estim)

        return fitted_parameter_vector



CONSTANTS = dict(
    water_diffusion_constant=2.299e-9,  # m^2/s
    water_in_axons_diffusion_constant=1.7e-9,  # m^2/s
    naa_in_axons=.00015e-9,  # m^2 / s
    water_gyromagnetic_ratio=267.513e6,   # 1/(sT)
)



# -*- coding: utf-8 -*-
import numpy as np
from dipy.data import get_sphere
from dipy.utils.optpkg import optional_package

SPHERE = get_sphere('symmetric362')
numba, have_numba, _ = optional_package("numba")

__all__ = [
    'perpendicular_vector',
    'rotation_matrix_around_100',
    'rotation_matrix_100_to_theta_phi',
    'rotation_matrix_100_to_xyz',
    'rotation_matrix_001_to_xyz',
    'rotation_matrix_100_to_theta_phi_psi',
    'T1_tortuosity',
    'parameter_equality',
    'define_shell_indices',
    'cart2sphere',
    'cart2mu',
    'R2mu_psi',
    'sphere2cart',
    'unitsphere2cart_1d',
    'unitsphere2cart_Nd'
]


def perpendicular_vector(v):
    """Returns a perpendicular vector to vector "v".

    Parameters
    ----------
    v : array, shape (3)
        normally Cartesian unit vector, but can also be any vector.

    Returns
    -------
    v_perp : array, shape (3)
        If v is unit vector, v_perp is a Cartesian unit vector perpendicular
        to v.
    """
    v_ = v * 1.
    if v_[1] == 0 and v_[2] == 0:
        if v[0] == 0:
            raise ValueError('zero vector')
        else:
            v_perp = np.cross(v_, [0, 1, 0])
            v_perp /= np.linalg.norm(v_perp)
            return v_perp
    v_perp = np.cross(v_, [1, 0, 0])
    v_perp /= np.linalg.norm(v_perp)
    return v_perp


def rotation_matrix_around_100(psi):
    """Generates a rotation matrix that rotates around the x-axis (1, 0, 0).

    Parameters
    ----------
    psi : float,
        euler angle [0, pi].

    Returns
    -------
    R : array, shape (3 x 3)
        Rotation matrix.
    """
    cos_psi = np.cos(psi)
    sin_psi = np.sin(psi)
    R = np.zeros((3, 3))
    R[0, 0] = 1.
    R[1, 1] = cos_psi
    R[1, 2] = -sin_psi
    R[2, 1] = sin_psi
    R[2, 2] = cos_psi
    return R


def rotation_matrix_100_to_theta_phi(theta, phi):
    """Generates a rotation matrix that rotates from the x-axis (1, 0, 0) to
    an other position on the unit sphere.

    Parameters
    ----------
    theta : float,
        inclination of polar angle of main angle mu [0, pi].
    phi : float,
        polar angle of main angle mu [-pi, pi].

    Returns
    -------
    R : array, shape (3 x 3)
        Rotation matrix.
    """
    x, y, z = unitsphere2cart_1d([theta, phi])
    return rotation_matrix_100_to_xyz(x, y, z)


def rotation_matrix_100_to_xyz(x, y, z):
    """Generates a rotation matrix that rotates from the x-axis (1, 0, 0) to
    an other position in Cartesian space.

    Parameters
    ----------
    x, y, z : floats,
        position in Cartesian space.

    Returns
    -------
    R : array, shape (3 x 3)
        Rotation matrix.
    """
    if x == 1.:
        if y == 0.:
            if z == 0.:
                return np.eye(3)
    y2 = y ** 2
    z2 = z ** 2
    yz = y * z
    R = np.array([[x, -y, -z],
                  [y, (x * y2 + z2) / (y2 + z2), ((x - 1) * yz) / (y2 + z2)],
                  [z, ((x - 1) * yz) / (y2 + z2), (y2 + x * z2) / (y2 + z2)]])
    return R


def rotation_matrix_001_to_xyz(x, y, z):
    """Generates a rotation matrix that rotates from the z-axis (0, 0, 1) to
    an other position in Cartesian space.

    Parameters
    ----------
    x, y, z : floats,
        position in Cartesian space.

    Returns
    -------
    R : array, shape (3 x 3)
        Rotation matrix.
    """
    if np.all(np.r_[x, y, z] == np.r_[0., 0., 1.]):
        return np.eye(3)
    x2 = x ** 2
    y2 = y ** 2
    xy = x * y
    R = np.array([[(y2 + x2 * z) / (x2 + y2), (xy * (z - 1)) / (x2 + y2), x],
                  [(xy * (z - 1)) / (x2 + y2), (x2 + y2 * z) / (x2 + y2), y],
                  [-x, -y, z]])
    return R


def rotation_matrix_100_to_theta_phi_psi(theta, phi, psi):
    """Generates a rotation matrix that rotates from the x-axis (1, 0, 0) to
    an other position in Cartesian space, and rotates about its axis.

    Parameters
    ----------
    theta : float,
        inclination of polar angle of main angle mu [0, pi].
    phi : float,
        polar angle of main angle mu [-pi, pi].
    psi : float,
        angle in radians of the bingham distribution around mu [0, pi].

    Returns
    -------
    R : array, shape (3 x 3)
        Rotation matrix.
    """
    R_100_to_theta_phi = rotation_matrix_100_to_theta_phi(theta, phi)
    R_around_100 = rotation_matrix_around_100(psi)
    return np.dot(R_100_to_theta_phi, R_around_100)


class T1_tortuosity:
    def __init__(self, S0_intra=1., S0_extra=1.):
        """Tortuosity model for perpendicular extra-axonal diffusivity
        [1, 2, 3].

        Parameters
        ----------
        S0_intra: float,
            S0 response of the tissue associated to the intra-cellular
            compartment. Default: 1.
        S0_extra: float,
            S0 response of the tissue associated to the extra-cellular
            compartment. Default: 1.

        References
        -------
        .. [1] Bruggeman, Von DAG. "Berechnung verschiedener physikalischer
            Konstanten von heterogenen Substanzen. I. Dielektrizitätskonstanten
            und Leitfähigkeiten der Mischkörper aus isotropen Substanzen."
            Annalen der physik 416.7 (1935): 636-664.
        .. [2] Sen et al. "A self-similar model for sedimentary rocks with
            application to the dielectric constant of fused glass beads."
            Geophysics 46.5 (1981): 781-795.
        .. [3] Szafer et al. "Theoretical model for water diffusion in
            tissues." Magnetic resonance in medicine 33.5 (1995): 697-712.
        """
        self.S0_intra = S0_intra
        self.S0_extra = S0_extra

    def __call__(self, lambda_par, vf_intra, vf_extra=None):
        """
        If vf_extra=None, then vf_intra must be a nested volume fraction, in
        the sense that
        E_bundle = vf_intra * E_intra + (1 - vf_intra) * E_extra,
        with
        vf_intra + (1 - vf_intra) = 1.
        The volume fractions are scaled according to the passed S0 responses.
        If both vf_intra and vf_extra are given, then they have to be
        normalized fractions, in the sense that vf_intra + vf_extra <= 1.

        Parameters
        ----------
        lambda_par : float,
            parallel diffusivity.
        vf_intra : float,
            intra-axonal volume fraction [0, 1].
        vf_extra : float, (optional)
            extra-axonal volume fraction [0, 1].

        Returns
        -------
        lambda_perp : float,
            Perpendicular diffusivity.
        """
        if vf_extra is None:
            vf_extra = 1. - vf_intra
        else:
            vf_intra = vf_intra / (vf_intra + vf_extra)

        f = ((vf_intra * self.S0_extra) /
             (vf_intra * self.S0_extra + vf_extra * self.S0_intra))
        return (1. - f) * lambda_par


def parameter_equality(param):
    "Function to force two model parameters to be equal in the optimization"
    return param


def fractional_parameter(fractional_parameter, parameter2):
    """
    Defines fractional parameter link where the original parameter ranges
    between 0-1 of parameter2.

    Parameters
    ----------
    fractional_parameter : string
        new parameter that ranges between 0 and 1.
    parameter2 : string
        parameter that is larger or equal than the original parameter
    """
    return fractional_parameter * parameter2


def define_shell_indices(bvals, b_value_ranges):
    """ Function to facilitate defining shell indices given some manual ranges
    in b-values. This function is useful as in practice the actual b-values may
    fluctuate slightly around the planned b-value. This information is needed
    by some models doing spherical convolutions or spherical means.
    CAUTION: If a data set has variations in pulse duration delta and/or pulse
    separation Delta, then different shells can have similar b-values. This
    means these shells may not be separable in b-value, and you'll have to do
    it manually.

    Parameters
    ----------
    bvals : 1D array of size (N_data),
        The b-values corresponding to the measured DWIs.
    b_value_ranges : 2D array of size (N_shells, 2)
        A list indicating for every shell the lower and upper b-value range.

    Returns
    -------
    shell_indices : 1D integer array of size (N_data),
        The shell indices corresponding to each DWI measurement. The index 0
        corresponds with the b0 measurements, while higher numbers indicate
        other shells. The numbers are ordered in the same order they are given
        in b_value_ranges.
    mean_shell_bvals : 1D array of size (N_shells),
        The mean b-value in each shell.
    """

    shell_indices = np.empty_like(bvals, dtype=int)
    mean_shell_bvals = np.zeros(len(b_value_ranges))
    shell_counter = 0
    dwi_counter = 0
    for b_range in b_value_ranges:
        lower_range = b_range[0]
        upper_range = b_range[1]
        shell_mask = np.all([bvals >= lower_range, bvals <= upper_range],
                            axis=0)
        shell_indices[shell_mask] = shell_counter
        mean_shell_bvals[shell_counter] = np.mean(bvals[shell_mask])
        dwi_counter += sum(shell_mask)
        shell_counter += 1
    if dwi_counter < len(bvals):
        msg = ("b_value_ranges only covered " + str(dwi_counter) +
               " out of " + str(len(bvals)) + " dwis")
        raise ValueError(msg)
    return shell_indices, mean_shell_bvals


def cart2sphere(cartesian_coordinates):
    """"Function to estimate spherical coordinates from cartesian coordinates
    according to wikipedia. Conforms with the dipy notation.

    Parameters
    ----------
    cartesian_coordinates : array of size (3) or (N x 3),
        array of cartesian coordinate vectors [x, y, z].

    Returns
    -------
    spherical_coordinates : array of size (3) or (N x 3),
        array of spherical coordinate vectors [r, theta, phi].
        range of theta [0, pi]. range of phi [-pi, pi].
    """
    if np.ndim(cartesian_coordinates) == 1:
        x, y, z = cartesian_coordinates
        r = np.sqrt(x ** 2 + y ** 2 + z ** 2)
        if r > 0:
            theta = np.arccos(z / r)
        else:
            theta = 0
        phi = np.arctan2(y, x)
        spherical_coordinates = np.r_[r, theta, phi]
    elif np.ndim(cartesian_coordinates) == 2:
        x, y, z = cartesian_coordinates.T
        r = np.sqrt(x ** 2 + y ** 2 + z ** 2)
        theta = np.arccos(z / r)
        theta = np.where(r > 0, theta, 0.)
        phi = np.arctan2(y, x)
        spherical_coordinates = np.c_[r, theta, phi]
    else:
        msg = "coordinates must be array of size 3 or N x 3."
        raise ValueError(msg)
    return spherical_coordinates


def cart2mu(xyz):
    """
    Function to estimate spherical coordinates from cartesian coordinates
    according to wikipedia. Conforms with the dipy notation.

    Parameters
    ----------
    cartesian_coordinates : array of size (3) or (N x 3),
        array of cartesian coordinate vectors [x, y, z].

    Returns
    -------
    spherical_coordinates : array of size (2) or (N x 2),
        array of spherical coordinate vectors [theta, phi].
        range of theta [0, pi]. range of phi [-pi, pi].
    """
    shape = xyz.shape[:-1]
    mu = np.zeros(np.r_[shape, 2])
    r = np.linalg.norm(xyz, axis=-1)
    mu[..., 0] = np.arccos(xyz[..., 2] / r)  # theta
    mu[..., 1] = np.arctan2(xyz[..., 1], xyz[..., 0])
    mu[r == 0] = 0, 0
    return mu


def R2mu_psi(R):
    """
    Function to estimate orientation mu and secondary orientation angle psi
    from a 3x3 rotation matrix. Can be given array of rotation matrices.

    Parameters
    ----------
    R : Array of size (N, 3, 3)
        rotation matrices that possibly can be estimated by DTI.

    Returns
    -------
    mu : array of size (N, 2),
        orientations in [theta, phi] angles
    psi : array of size (N),
        secondary orientation psi (for Bingham for example).
    """
    mu = cart2mu(R[..., :, 0])
    mu_flat = mu.reshape([-1, 2])
    R_flat = R.reshape([-1, 3, 3])
    psi_flat = np.zeros(len(mu_flat))
    for i in range(len(mu_flat)):
        R_theta_phi = rotation_matrix_100_to_theta_phi(
            mu_flat[i, 0], mu_flat[i, 0])
        psi_flat[i] = np.arcsin(np.dot(R_theta_phi.T, R_flat[i])[2, 1])
    psi = psi_flat.reshape(mu.shape[:-1])
    psi[psi < 0] += np.pi
    return mu, psi


def sphere2cart(spherical_coordinates):
    """"Function to estimate cartesian coordinates from spherical coordinates
    according to wikipedia. Conforms with the dipy notation.

    Parameters
    ----------
    spherical_coordinates : array of size (3) or (N x 3),
        array of spherical coordinate vectors [r, theta, phi].
        range of theta [0, pi]. range of phi [-pi, pi].

    Returns
    -------
    cartesian_coordinates : array of size (3) or (N x 3),
        array of cartesian coordinate vectors [x, y, z].
    """
    if np.ndim(spherical_coordinates) == 1:
        r, theta, phi = spherical_coordinates
        sintheta = np.sin(theta)
        x = r * sintheta * np.cos(phi)
        y = r * sintheta * np.sin(phi)
        z = r * np.cos(theta)
        cartesian_coordinates = np.r_[x, y, z]
    elif np.ndim(spherical_coordinates) == 2:
        r, theta, phi = spherical_coordinates.T
        sintheta = np.sin(theta)
        x = r * sintheta * np.cos(phi)
        y = r * sintheta * np.sin(phi)
        z = r * np.cos(theta)
        cartesian_coordinates = np.c_[x, y, z]
    else:
        msg = "coordinates must be array of size 3 or N x 3."
        raise ValueError(msg)
    return cartesian_coordinates


def unitsphere2cart_1d(mu):
    """Optimized function dedicated to convert 1D unit sphere coordinates
    to cartesian coordinates.

    Parameters
    ----------
    mu : array of size (2)
        unit sphere coordinates, as theta, phi = mu

    Returns
    -------
    mu_cart, array of size (3)
        mu in cartesian coordinates, as x, y, z = mu_cart
    """
    theta, phi = mu
    mu_cart = np.zeros(3)
    sintheta = np.sin(theta)
    mu_cart[0] = sintheta * np.cos(phi)
    mu_cart[1] = sintheta * np.sin(phi)
    mu_cart[2] = np.cos(theta)
    return mu_cart


def unitsphere2cart_Nd(mu):
    """Optimized function deicated to convert 1D unit sphere coordinates
    to cartesian coordinates.

    Parameters
    ----------
    mu : Nd array of size (..., 2)
        unit sphere coordinates, as theta, phi = mu

    Returns
    -------
    mu_cart, Nd array of size (..., 3)
        mu in cartesian coordinates, as x, y, z = mu_cart
    """
    mu_cart = np.zeros(np.r_[mu.shape[:-1], 3])
    theta = mu[..., 0]
    phi = mu[..., 1]
    sintheta = np.sin(theta)
    mu_cart[..., 0] = sintheta * np.cos(phi)
    mu_cart[..., 1] = sintheta * np.sin(phi)
    mu_cart[..., 2] = np.cos(theta)
    return mu_cart


if have_numba:
    unitsphere2cart_1d = numba.njit()(unitsphere2cart_1d)



import numpy as np
from .constants import CONSTANTS


__all__ = [
    'q_from_b',
    'b_from_q',
    'q_from_g',
    'g_from_q',
    'b_from_g',
    'g_from_b'
]


def q_from_b(
    b, delta, Delta
):
    """Compute q-value from b-value. Units are standard units."""
    tau = Delta - delta / 3
    q = np.sqrt(b / tau) / (2 * np.pi)
    return q


def b_from_q(
    q, delta, Delta
):
    """Compute b-value from q-value. Units are standard units."""
    tau = Delta - delta / 3
    b = (q * (2 * np.pi)) ** 2 * tau
    return b


def q_from_g(
    g, delta,
    gyromagnetic_ratio=CONSTANTS['water_gyromagnetic_ratio']
):
    """Compute q-value from gradient strength. Units are standard units."""
    q = g * delta * gyromagnetic_ratio / (2 * np.pi)
    return q


def g_from_q(
    q, delta,
    gyromagnetic_ratio=CONSTANTS['water_gyromagnetic_ratio']
):
    """Compute gradient strength from q-value. Units are standard units."""
    return q * (2 * np.pi) / (delta * gyromagnetic_ratio)


def b_from_g(
    g, delta, Delta,
    gyromagnetic_ratio=CONSTANTS['water_gyromagnetic_ratio']
):
    """Compute b-value from gradient strength. Units are standard units."""
    tau = Delta - delta / 3
    return (g * gyromagnetic_ratio * delta) ** 2 * tau


def g_from_b(
    b, delta, Delta,
    gyromagnetic_ratio=CONSTANTS['water_gyromagnetic_ratio']
):
    """Compute gradient strength from b-value. Units are standard units."""
    tau = Delta - delta / 3
    return np.sqrt(b / tau) / (gyromagnetic_ratio * delta)



import pathlib
import boto
import pkg_resources
import os
import nibabel as nib
import numpy as np


DATA_PATH = pkg_resources.resource_filename(
    'dmipy', 'data'
)

__all__ = [
    'HCPInterface'
]


class HCPInterface:
    """
    Function to download HCP data directly from AWS server.
    Since HCP requires you to sign its user agreements, this function
    takes as input the user's personal public and secret AWS key.

    These keys can be generated following the the instructions here:
    https://wiki.humanconnectome.org/display/PublicData/How+To+Connect+to+Connectome+Data+via+AWS

    The reason this function is set up this way is because we cannot sign
    the user agreement for the user. Now, the user takes its own responsibility
    and we can provide this convenient downloader without having to go through
    the HCP interface.

    Parameters
    ----------
    your_aws_public_key : string
        user aws public key
    your_aws_secret_key : string
        user aws secret key
    """

    def __init__(self, your_aws_public_key, your_aws_secret_key):
        self.public_key = your_aws_public_key
        self.secret_key = your_aws_secret_key

        s3 = boto.connect_s3(
            aws_access_key_id=self.public_key,
            aws_secret_access_key=self.secret_key,
        )

        for key in s3.get_all_buckets():
            if key.name == 'hcp-openaccess':
                self.s3_bucket = key
                break

        self.hcp_directory = os.path.join(DATA_PATH, 'hcp')
        if not os.path.exists(self.hcp_directory):
            os.makedirs(self.hcp_directory)

    def download_and_prepare_dmipy_example_dataset(self):
        """
        Downloads subect 100307 of the Wu-Minn HCP data and prepares it to be
        used for the dmipy example notebooks.
        """
        subject_ID = 100307
        self.download_subject(subject_ID)
        self.prepare_example_slice(subject_ID)

    @property
    def available_subjects(self):
        subjects_textfile = os.path.join(DATA_PATH, "wu_minn_hcp_subjects.txt")
        return np.loadtxt(subjects_textfile, dtype=int)

    def download_subject(self, subject_ID):
        """
        Downloads Wu-Minn HCP subject data to the dmipy data folder.
        The downloaded data includes the b-values, gradient orientations,
        diffusion-weighted images and the binary brain mask.

        Parameters
        ----------
        subject_ID: integer
            the identification number of the Wu-Minn HCP subject
        """
        hcp_data_path = os.path.join(self.hcp_directory, str(subject_ID))

        if not os.path.exists(hcp_data_path):
            os.makedirs(hcp_data_path)

        print('Downloading data to {}'.format(hcp_data_path))

        counter = 0
        for key in self.s3_bucket.list("HCP_1200"):
            path = pathlib.Path(key.name)
            if (
                len(path.parts) == 5 and
                subject_ID == int(path.parts[1]) and
                path.parts[-2] == "Diffusion"
            ):
                if (
                    'bval' in path.parts[-1] or
                    'bvec' in path.parts[-1] or
                    'data' in path.parts[-1] or
                    'nodif' in path.parts[-1]
                ):
                    print('Downloading {}'.format(path.parts[-1]))
                    filepath = os.path.join(hcp_data_path, path.parts[-1])
                    with open(filepath, 'wb') as f:
                        key.get_contents_to_file(f)
                    counter += 1
                    if counter == 4:
                        break

    def prepare_example_slice(self, subject_ID):
        "Prepares a coronal slice for the dmipy example notebooks."
        msg = "Preparing coronal slice for dmipy examples"
        print(msg)

        folder_name = "hcp_example_slice"
        example_directory = os.path.join(self.hcp_directory, folder_name)
        if not os.path.exists(example_directory):
            os.makedirs(example_directory)

        subject_data_path = os.path.join(self.hcp_directory, str(subject_ID))

        data = nib.load(os.path.join(
            subject_data_path, 'data.nii.gz')).get_data()
        affine = nib.load(os.path.join(
            subject_data_path, 'data.nii.gz')).affine
        mask = nib.load(os.path.join(
            subject_data_path, 'nodif_brain_mask.nii.gz')).get_data()
        data_shape = data.shape

        slice_index = data_shape[1] // 2
        data_slice = data[:, slice_index: slice_index + 1]
        mask_slice = mask[:, slice_index: slice_index + 1]
        data_slice[mask_slice == 0] = 0

        nib.save(nib.Nifti1Image(data_slice, affine), os.path.join(
            example_directory, 'coronal_slice.nii.gz'))
        print('Done')



from __future__ import absolute_import, division, print_function
from os.path import join as pjoin

# Format expected by setup.py and doc/source/conf.py: string of form "X.Y.Z"
_version_major = 1
_version_minor = 0
_version_micro = 5  # use '' for first of series, number for 1 and above
_version_extra = ''
# _version_extra = ''  # Uncomment this for full releases

# Construct full version string from these.
_ver = [_version_major, _version_minor]
if _version_micro:
    _ver.append(_version_micro)
if _version_extra:
    _ver.append(_version_extra)

__version__ = '.'.join(map(str, _ver))

CLASSIFIERS = ["Development Status :: 3 - Alpha",
               "Environment :: Console",
               "Intended Audience :: Science/Research",
               "License :: OSI Approved :: MIT License",
               "Operating System :: OS Independent",
               "Programming Language :: Python",
               "Topic :: Scientific/Engineering"]

# Description should be a one-liner:
description = "dmipy: diffusion microstructure imaging in python"
# Long description will go up on the pypi page
long_description_content_type = 'text/markdown'

NAME = "dmipy"
MAINTAINER = "Rutger Fick"
MAINTAINER_EMAIL = "fick.rutger@gmail.com"
DESCRIPTION = description
LONG_DESCRIPTION_CONTENT_TYPE = long_description_content_type
URL = "https://github.com/AthenaEPI/dmipy"
DOWNLOAD_URL = ""
LICENSE = "MIT"
AUTHOR = "Rutger Fick"
AUTHOR_EMAIL = "fick.rutger@gmail.com"
PLATFORMS = "OS Independent"
MAJOR = _version_major
MINOR = _version_minor
MICRO = _version_micro
VERSION = __version__
PACKAGE_DATA = {
    'dmipy': [
        pjoin('data', 'gradient_tables', '*'),
        pjoin('data', 'spheres', '*'),
        pjoin('data', 'bingham_normalization_splinefit.npz'),
        pjoin('data', 'wu_minn_hcp_subjects.txt'),
        pjoin('data', 'wu_minn_hcp_coronal_slice.nii.gz'),
        pjoin('data', 'tanguy_cat_spinal_cord', '*'),
        pjoin('data', 'camino', '*'),
        pjoin('data', 'de_santis_camino', '*'),
        pjoin('data', 'isbi2015_white_matter_challenge', '*')
    ]
}
REQUIRES = ["dipy", "scipy (==1.4.1)", "numpy (>=1.13)"]



from scipy.optimize import brute, minimize
import pkg_resources
from ..utils.utils import cart2mu
import numpy as np
from dipy.utils.optpkg import optional_package

numba, have_numba, _ = optional_package("numba")

SPHERES_PATH = pkg_resources.resource_filename(
    'dmipy', 'data/spheres'
)

__all__ = [
    'GlobalBruteOptimizer',
    'Brute2FineOptimizer',
    'nested_to_normalized_fractions',
    'normalized_to_nested_fractions_array',
    'find_minimum_argument'
]


class GlobalBruteOptimizer:
    r"""
    Brute-Force optimizer. Given a model and an acquisition scheme, first
    computes a global grid of parameters and corresponding signal attenuations.
    All parameters except the spherical orientation parameter 'mu' are
    sampled between their corresponding parameter_ranges in 'Ns' equal steps.
    For 'mu' a spherical grid of 'N_sphere_samples" points is used, which were
    generated using the work of Caruyer et al. [1].

    When calling the function with measured data, the closest parameters are
    return based on the sum-squared error between the signal grid and the data.

    Parameters
    ----------
    model: dmipy MultiCompartmentModel instance,
        Can be composed of any model combination.
    acquisition_scheme: DmipyAcquisitionScheme instance,
        acquisition scheme of the to-be-fitted data.
    x0_vector: array of size (Nparameters,)
        optional initial guess parameters. As long as the initial guess does
        not vary voxel-by-voxel the parameter grid will be estimated, only
        including the initial guess value for the parameters that were included
        in x0_vector.
    Ns: integer,
        number of equally spaced sampling points for regular parameters.
    N_sphere_sampled: integer,
        number of sampled sphere points to sample orientation 'mu'.

    References
    ----------
    .. [1] Caruyer, Emmanuel, et al. "Design of multishell sampling schemes
        with uniform coverage in diffusion MRI." Magnetic resonance in medicine
        69.6 (2013): 1534-1540.
    """

    def __init__(self, model, acquisition_scheme,
                 x0_vector=None, Ns=5, N_sphere_samples=30):
        self.model = model
        self.acquisition_scheme = acquisition_scheme
        self.x0_vector = x0_vector
        self.Ns = Ns

        if x0_vector is None:
            self.global_optimization_grid = True
            x0_vector = np.tile(np.nan, len(model.bounds_for_optimization))
            self.precompute_signal_grid(model, x0_vector, Ns, N_sphere_samples)
        elif x0_vector.squeeze().ndim == 1 and np.any(np.isnan(x0_vector)):
            self.global_optimization_grid = True
            self.precompute_signal_grid(
                model, x0_vector.squeeze(), Ns, N_sphere_samples)
        elif np.all(np.isnan(x0_vector.reshape([-1, x0_vector.shape[-1]])[0])):
            x0_vector_ = np.tile(np.nan, len(model.bounds_for_optimization))
            self.global_optimization_grid = True
            self.precompute_signal_grid(
                model, x0_vector_, Ns, N_sphere_samples)
        else:
            self.global_optimization_grid = False
            msg = "Cannot estimate signal grid with voxel-dependent x0_vector."
            print(msg)

    def precompute_signal_grid(self, model, x0_vector, Ns, N_sphere_samples):
        """
        Function that estimates the parameter grid and corresponding signal
        attenuation.

        NOTE: In the current implementation initial guesses for volume
        fractions are still ignored
        """

        # import sphere array mu as (theta, phi)
        sphere_vertices = np.loadtxt(
            SPHERES_PATH + "/01-shells-" + str(N_sphere_samples) + ".txt",
            skiprows=1)[:, 1:]
        mu = cart2mu(sphere_vertices)
        grids_per_mu = []
        N_model_fracts = 0
        parameter_cardinality_items = list(model.parameter_cardinality.items())
        if self.model.N_models > 1:
            N_model_fracts = self.model.N_models
            parameter_cardinality_items = parameter_cardinality_items[
                :-N_model_fracts
            ]

        max_cardinality = np.max(list(model.parameter_cardinality.values()))
        for card_counter in range(max_cardinality):
            per_parameter_vectors = []
            counter = 0
            for name, card in parameter_cardinality_items:
                par_range = model.parameter_ranges[name]
                if card == 1:
                    if np.isnan(x0_vector[counter]):
                        per_parameter_vectors.append(np.linspace(
                            par_range[0], par_range[1], Ns) *
                            model.parameter_scales[name])
                    else:
                        per_parameter_vectors.append([x0_vector[counter]])
                    counter += 1
                if card == 2:
                    if np.isnan(x0_vector[counter]):
                        per_parameter_vectors.append(
                            mu[:, card_counter] *
                            model.parameter_scales[name][0])
                    else:
                        per_parameter_vectors.append(
                            [x0_vector[counter + card_counter]])
                    per_parameter_vectors.append(np.nan)
                    counter += 2
            # append nested volume fractions now.
            if self.model.N_models > 1:
                for _ in range(N_model_fracts - 1):
                    per_parameter_vectors.append(np.linspace(0., 1., Ns))
            grids_per_mu.append(np.meshgrid(*per_parameter_vectors))

        counter = 0
        param_dict = {}
        for name, card in parameter_cardinality_items:
            if card == 1:
                param_dict[name] = grids_per_mu[0][counter].reshape(-1)
                counter += 1
            if card == 2:
                param_dict[name] = np.concatenate(
                    [grids_per_mu[0][counter][..., None],
                     grids_per_mu[1][counter][..., None]], axis=-1).reshape(
                    [-1, 2])
                counter += 2

        # now add nested to regular volume fractions
        if self.model.N_models > 1:
            nested_fractions = grids_per_mu[0][-(N_model_fracts - 1):]
            lin_nested_fractions = [
                fracts.reshape(-1) for fracts in nested_fractions]
            lin_fractions = np.empty(
                (len(lin_nested_fractions[0]), N_model_fracts))
            for i in range(len(lin_nested_fractions[0])):
                lin_fractions[i] = nested_to_normalized_fractions(
                    np.r_[[fract[i] for fract in lin_nested_fractions]])

            counter = 0
            for name, card in list(model.parameter_cardinality.items())[
                    -N_model_fracts:]:
                param_dict[name] = lin_fractions[:, counter]
                counter += 1

        self.parameter_grid = model.parameters_to_parameter_vector(
            **param_dict)
        self.signal_grid = model.simulate_signal(
            self.acquisition_scheme, self.parameter_grid)

    def __call__(self, data, parameter_scale_normalization=True):
        """
        Calculates the closest parameter combination based on the sum-squared
        distances between the measured data and the simulated signal grid.

        Parameters
        ----------
        data: array of size (Ndata,),
            array containing the measured signal attenuation.
        parameter_scale_normalization: bool,
            option whether to return the model parameters in their 'true' scale
            or in the normalized scale where they are all around scale O(1).

        Returns
        -------
        parameters_brute: array of size (Nparameters,),
            estimated closest model parameters in the parameter grid.
        """
        if self.global_optimization_grid is True:
            argmin = find_minimum_argument(self.signal_grid, data)
            parameters_brute = self.parameter_grid[argmin]
            if parameter_scale_normalization:
                return parameters_brute / self.model.scales_for_optimization
            return parameters_brute
        else:
            msg = "Global Parameter Grid could not be set because parameter "
            msg += "initial condition is voxel dependent."
            raise ValueError(msg)


class Brute2FineOptimizer:
    """
    Brute force optimizer with refining. Essentially this function does both
    the brute force optimization like GlobalBruteOptimizer (without treating
    mu differently, which is currently suboptimal), but then follows it with
    a gradient-descent based refining step [1, 2] to find the local minimum.

    All parameters are optimized within their parameter_bounds. Volume
    fraction are optimized by 'nesting' them, such that given a set of models
    :math:`m_1...m_N`, and the partial volume ratios math:`v_1...v_{N-1}`, the
    partial volume function is

    .. math::
        v_1 m_1 + (1 - v_1) v_2 m_2 + ... + (1 - v_1)...(1-v_{N-1}) m_N

    Parameters
    ----------
    model: dmipy MultiCompartmentModel instance,
        Can be composed of any model combination.
    acquisition_scheme: DmipyAcquisitionScheme instance,
        acquisition scheme of the to-be-fitted data.
    Ns: integer,
        number of equally spaced sampling points for regular parameters.

    References
    ----------
    .. [1] Byrd, R H and P Lu and J. Nocedal. 1995. A Limited Memory Algorithm
        for Bound Constrained Optimization. SIAM Journal on Scientific and
        Statistical Computing 16 (5): 1190-1208.
    .. [2] Zhu, C and R H Byrd and J Nocedal. 1997. L-BFGS-B: Algorithm 778:
        L-BFGS-B, FORTRAN routines for large scale bound constrained
        optimization. ACM Transactions on Mathematical Software 23 (4):
        550-560.
    """

    def __init__(self, model, acquisition_scheme, Ns=5):
        self.model = model
        self.acquisition_scheme = acquisition_scheme
        self.Ns = Ns
        if model.N_models > 1 and model.volume_fractions_fixed:
            self.obj_func = self.objective_function_vf_fixed
        else:
            self.obj_func = self.objective_function

    def objective_function(self, parameter_vector, data):
        "The objective function for brute-force and gradient-based optimizer."
        if self.model.N_models > 1:
            nested_fractions = parameter_vector[-(self.model.N_models - 1):]
            normalized_fractions = nested_to_normalized_fractions(
                nested_fractions)
            parameter_vector_ = np.r_[
                parameter_vector[:-(self.model.N_models - 1)],
                normalized_fractions]
        else:
            parameter_vector_ = parameter_vector
        parameter_vector_ = (
            parameter_vector_ * self.model.scales_for_optimization)
        parameters = {}
        parameters.update(
            self.model.parameter_vector_to_parameters(parameter_vector_)
        )
        E_model = self.model(self.acquisition_scheme, **parameters)
        E_diff = E_model - data
        objective = np.dot(E_diff, E_diff) / len(data)
        return objective

    def objective_function_vf_fixed(self, parameter_vector, data, vf):
        "The objective function if the volume fractions have been fixed."
        parameter_vector_ = np.hstack([parameter_vector, vf])
        parameter_vector_ = (
            parameter_vector_ * self.model.scales_for_optimization)
        parameters = {}
        parameters.update(
            self.model.parameter_vector_to_parameters(parameter_vector_)
        )
        E_model_sep = self.model(
            self.acquisition_scheme, quantity="stochastic cost function",
            **parameters)
        E_model = np.dot(E_model_sep, vf)
        E_diff = E_model - data
        objective = np.dot(E_diff, E_diff) / len(data)
        return objective

    def __call__(self, data, x0_vector):
        """
        Estimates the model parameters given the measured signal attenuation
        and an initial parameter guess. For parameters that are not given an
        initial guess, we find a good first guess using a brute-force approach.

        Parameters
        ----------
        data: array of size (Ndata,),
            array containing the measured signal attenuation.
        x0_vector: array of size (Nparameters,),
            initial guess array that either contains a float for parameters
            with an initial guess or None for parameters that have no guess.

        Returns
        -------
        x_fine: array of size (Nparameters,),
            array of the optimized model parameters.
        """
        bounds = self.model.bounds_for_optimization
        bounds_brute = []
        bounds_fine = list(bounds)
        for i, x0_ in enumerate(x0_vector):
            if np.isnan(x0_):
                bounds_brute.append(
                    slice(bounds[i][0], bounds[i][1],
                          (bounds[i][1] - bounds[i][0]) / float(self.Ns)))
            if not np.isnan(x0_):
                bounds_brute.append(slice(x0_, x0_ + 1e-2, None))
            if (not np.isnan(x0_) and
                    self.model.opt_params_for_optimization[i] is False):
                bounds_fine[i] = np.r_[x0_, x0_]

        if self.model.N_models > 1 and not self.model.volume_fractions_fixed:
            # go to nested bounds
            bounds_brute = bounds_brute[:-1]
            bounds_fine = bounds_fine[:-1]
            x0_vector = x0_vector[:-1]
            fit_args = (data,)
        elif self.model.N_models > 1 and self.model.volume_fractions_fixed:
            # separate nonlinear and linear parameters
            vfs = x0_vector[-self.model.N_models:]
            bounds_brute = bounds_brute[:-self.model.N_models]
            bounds_fine = bounds_fine[:-self.model.N_models]
            x0_vector = x0_vector[:-self.model.N_models]
            fit_args = (data, vfs)
        else:
            fit_args = (data,)

        if np.any(np.isnan(x0_vector)):
            x0_brute = brute(
                self.obj_func, ranges=bounds_brute, args=fit_args,
                finish=None)
        else:
            x0_brute = x0_vector

        x_fine_nested = minimize(self.obj_func, x0_brute,
                                 args=fit_args, bounds=bounds_fine,
                                 method='L-BFGS-B').x
        if self.model.N_models > 1 and not self.model.volume_fractions_fixed:
            nested_fractions = x_fine_nested[-(self.model.N_models - 1):]
            normalized_fractions = nested_to_normalized_fractions(
                nested_fractions)
            x_fine = np.r_[
                x_fine_nested[:-(self.model.N_models - 1)],
                normalized_fractions]
        elif self.model.N_models > 1 and self.model.volume_fractions_fixed:
            x_fine = np.hstack([x_fine_nested, vfs])
        else:
            x_fine = x_fine_nested
        return x_fine


def nested_to_normalized_fractions(nested_fractions):
    "Calculates the normal volume fractions from nested ones."
    N = len(nested_fractions)
    normalized_fractions = np.zeros(N + 1)
    remaining_fraction = 1.
    for i in range(N):
        normalized_fractions[i] = remaining_fraction * nested_fractions[i]
        remaining_fraction -= normalized_fractions[i]
    normalized_fractions[-1] = remaining_fraction
    return normalized_fractions


def normalized_to_nested_fractions_array(normalized_fractions):
    "Calculates the nested volume fractions from normal ones."
    norm_fracts = np.atleast_2d(normalized_fractions)
    N = norm_fracts.shape[-1]
    nested_fractions = np.zeros(np.r_[norm_fracts.shape[:-1], N - 1])
    remaining_fraction = np.ones(norm_fracts.shape[:-1])
    for i in range(N - 1):
        nested_fractions[..., i] = normalized_fractions[...,
                                                        i] / remaining_fraction
        remaining_fraction -= normalized_fractions[..., i]
    return nested_fractions


if have_numba:
    @numba.njit()
    def find_minimum_argument(data_grid, signal):
        cost = np.zeros(len(data_grid))
        for i in range(len(data_grid)):
            diff = data_grid[i] - signal
            cost[i] = np.dot(diff, diff)
        return np.argmin(cost)
else:
    def find_minimum_argument(data_grid, signal):
        """
        Finds the index in the simulated data_grid that has the
        lowest sum-squared error to the signal.
        """
        return np.argmin(np.sum((data_grid - signal) ** 2, axis=-1))



from dipy.reconst import dti
from ..core.acquisition_scheme import gtab_dmipy2dipy
from ..core.modeling_framework import (
    MultiCompartmentSphericalHarmonicsModel)
import numpy as np
from dipy.segment.mask import median_otsu
from dipy.data import get_sphere, HemiSphere
from ..signal_models.tissue_response_models import (
    estimate_TR2_anisotropic_tissue_response_model)
from scipy.ndimage import binary_erosion


def white_matter_response_tournier07(
        acquisition_scheme, data, N_candidate_voxels=300, **kwargs):
    """The original white matter response estimation algorithm according to
    [1]_. In essence, it just takes the 300 voxels with the highest FA, aligns
    them with the z-axis, and estimates the averaged white matter response from
    that.

    Parameters
    ----------
    acquisition_scheme : DmipyAcquisitionScheme instance,
        An acquisition scheme that has been instantiated using dMipy.
    data : NDarray,
        Measured diffusion signal array.

    Returns
    -------
    S0_wm : positive float,
        Estimated S0 tissue response value.
    TR2_wm_model : Dmipy Anisotropic ModelFree Model
        ModelFree representation of white matter response.
    selected_indices : array of size (N_candidate_voxels,),
        indices of selected voxels for white matter response.

    References
    ----------
    .. [1] Tournier, J-Donald, Fernando Calamante, and Alan Connelly. "Robust
        determination of the fibre orientation distribution in diffusion MRI:
        non-negativity constrained super-resolved spherical deconvolution."
        Neuroimage 35.4 (2007): 1459-1472.
    """
    data_shape = np.atleast_2d(data).shape
    N_voxels = int(np.prod(data_shape[:-1]))
    if N_voxels < N_candidate_voxels:
        msg = "The original algorithm uses 300 candidate voxels to estimate "
        msg += "the tissue response. Currently only {} ".format(N_voxels)
        msg += "candidate voxels given."
        print(msg)
        N_candidate_voxels = N_voxels

    if data.ndim == 4:
        # calculate brain mask on 4D data (x, y, z, DWI)
        b0_mask, mask = median_otsu(
            input_volume=data,
            vol_idx=np.where(acquisition_scheme.b0_mask)[0],
            median_radius=4, numpass=4)  # based on dipy default
        # needs to be eroded 3 times.
        mask_eroded = binary_erosion(mask, iterations=3)
        data_to_fit = data[mask_eroded]
    else:
        # can't calculate brain mask on other than 4D data.
        # assume the data was prepared.
        data_to_fit = data.reshape([-1, data_shape[-1]])

    gtab = gtab_dmipy2dipy(acquisition_scheme)
    tenmod = dti.TensorModel(gtab)
    tenfit = tenmod.fit(data_to_fit)
    fa = tenfit.fa

    # selected based on FA
    selected_indices = np.argsort(fa)[-N_candidate_voxels:]
    selected_data = data_to_fit[selected_indices]
    S0_wm, TR2_wm_model = estimate_TR2_anisotropic_tissue_response_model(
        acquisition_scheme, selected_data)
    return S0_wm, TR2_wm_model, selected_indices


def white_matter_response_tournier13(
        acquisition_scheme, data, max_iter=5, sh_order=10,
        N_candidate_voxels=300, peak_ratio_setting='mrtrix'):
    """
    Iterative model-free white matter response function estimation according to
    [1]_. Quoting the paper, the steps are the following:

    - 1) The 300 brain voxels with the highest FA were identified within a
        brain mask (eroded by three voxels to remove any noisy voxels at the
        brain edges).
    - 2) The single-fibre 'response function' was estimated within these
        voxels, and used to compute the fibre orientation distribution (FOD)
        employing constrained spherical deconvolution (CSD) up to lmax = 10.
    - 3) Within each voxel, a peak-finding procedure was used to identify the
        two largest FOD peaks, and their amplitude ratio was computed.
    - 4) The 300 voxels with the lowest second to first peak amplitude ratios
        were identified, and used as the current estimate of the set of
        'single-fibre' voxels. It should be noted that these voxels were not
        required to be a subset of the original set of 'single-fibre' voxels.
    - 5) To ensure minimal bias from the initial estimate of the 'response
        function', steps (2) to (4) were re-iterated until convergence (no
        difference in the set of 'single-fibre' voxels). It should be noted
        that, in practice, convergence was achieved within a single iteration
        in all cases.

    Parameters
    ----------
    acquisition_scheme : DmipyAcquisitionScheme instance,
        An acquisition scheme that has been instantiated using dMipy.
    data : NDarray,
        Measured diffusion signal array.
    max_iter : Positive integer,
        Defines the maximum amount of iterations to be done for the single-
        fibre response kernel.
    sh_order : Positive even integer,
        Maximum spherical harmonics order to be used in the FOD estimation for
        the single-fibre response kernel.
    N_candidate_voxels : integer,
        Number of voxels to be included in the final white matter response
        estimation. Default is 300 following [1]_.
    peak_ratio_setting : string,
        Can be either 'ratio' or 'mrtrix', meaning the 'ratio' parameter
        between two peaks is actually calculated as the ratio, or a more
        complicated version as 1 / sqrt(peak1 * (1 - peak2 / peak1)) ** 2, to
        avoid favouring small, yet low SNR FODs [2]_.

    Returns
    -------
    S0_wm : positive float,
        Estimated S0 tissue response value.
    TR2_wm_model : Dmipy Anisotropic ModelFree Model
        ModelFree representation of white matter response.
    selected_indices : array of size (N_candidate_voxels,),
        indices of selected voxels for white matter response.

    References
    ----------
    .. [1] Tournier, J-Donald, Fernando Calamante, and Alan Connelly.
        "Determination of the appropriate b value and number of gradient
        directions for high-angular-resolution diffusion-weighted imaging."
        NMR in Biomedicine 26.12 (2013): 1775-1786.
    .. [2] MRtrix 3.0 readthedocs
    """
    data_shape = np.atleast_2d(data).shape
    N_voxels = int(np.prod(data_shape[:-1]))
    if N_voxels < N_candidate_voxels:
        msg = "The parameter N_candidate voxels is set to {} but only ".format(
            N_candidate_voxels)
        msg += "{} voxels are given. N_candidate_voxels".format(N_voxels)
        msg += " reset to number of voxels given."
        print(msg)
        N_candidate_voxels = N_voxels

    ratio_settings = ['ratio', 'mrtrix']
    if peak_ratio_setting not in ratio_settings:
        msg = 'peak_ratio_setting must be in {}'.format(ratio_settings)
        raise ValueError(msg)

    if data.ndim == 4:
        # calculate brain mask on 4D data (x, y, z, DWI)
        b0_mask, mask = median_otsu(
            input_volume=data,
            vol_idx=np.where(acquisition_scheme.b0_mask)[0],
            median_radius=4, numpass=4)  # based on dipy default
        # needs to be eroded 3 times.
        mask_eroded = binary_erosion(mask, iterations=3)
        data_to_fit = data[mask_eroded]
    else:
        # can't calculate brain mask on other than 4D data.
        # assume the data was prepared.
        data_to_fit = data.reshape([-1, data_shape[-1]])

    gtab = gtab_dmipy2dipy(acquisition_scheme)
    tenmod = dti.TensorModel(gtab)
    tenfit = tenmod.fit(data_to_fit)
    fa = tenfit.fa

    # selected based on FA
    selected_indices = np.argsort(fa)[-N_candidate_voxels:]
    sphere = get_sphere('symmetric724')
    hemisphere = HemiSphere(theta=sphere.theta, phi=sphere.phi)
    # iterate until convergence
    it = 0
    while True:
        print('Tournier13 white matter response iteration {}'.format(it + 1))
        selected_data = data_to_fit[selected_indices]

        S0_wm, TR2_wm_model = estimate_TR2_anisotropic_tissue_response_model(
            acquisition_scheme, selected_data)
        sh_model = MultiCompartmentSphericalHarmonicsModel(
            [TR2_wm_model], sh_order=sh_order)
        sh_fit = sh_model.fit(acquisition_scheme, data_to_fit,
                              solver='csd_tournier07',
                              use_parallel_processing=False,
                              lambda_lb=0.)
        peaks, values, indices = sh_fit.peaks_directions(
            hemisphere, max_peaks=2, relative_peak_threshold=0.)
        if peak_ratio_setting == 'ratio':
            ratio = values[..., 1] / values[..., 0]
        elif peak_ratio_setting == 'mrtrix':
            ratio = 1. / np.sqrt(
                values[..., 0] * (1 - values[..., 1] / values[..., 0])) ** 2
        selected_indices_old = selected_indices
        selected_indices = np.argsort(ratio)[:N_candidate_voxels]
        percentage_overlap = 100 * float(len(np.intersect1d(
            selected_indices, selected_indices_old))) / N_candidate_voxels
        print('{:.1f} percent candidate voxel overlap.'.format(
            percentage_overlap))
        if percentage_overlap == 100.:
            print('White matter response converged')
            break
        it += 1
        if it == max_iter:
            print('Maximum iterations reached without convergence')
            break
    return S0_wm, TR2_wm_model, selected_indices
